# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# pyre-unsafe

from typing import Any, Iterable, Mapping, Sequence

from libcst._parser.base_parser import BaseParser
from libcst._parser.grammar import get_nonterminal_conversions, get_terminal_conversions
from libcst._parser.parso.pgen2.generator import Grammar
from libcst._parser.parso.python.token import TokenType
from libcst._parser.types.config import ParserConfig
from libcst._parser.types.conversions import NonterminalConversion, TerminalConversion
from libcst._parser.types.token import Token


class PythonCSTParser(BaseParser[Token, TokenType, Any]):
    config: ParserConfig
    terminal_conversions: Mapping[str, TerminalConversion]
    nonterminal_conversions: Mapping[str, NonterminalConversion]

    def __init__(
        self,
        *,
        tokens: Iterable[Token],
        config: ParserConfig,
        pgen_grammar: "Grammar[TokenType]",
        start_nonterminal: str = "file_input",
    ) -> None:
        super().__init__(
            tokens=tokens,
            lines=config.lines,
            pgen_grammar=pgen_grammar,
            start_nonterminal=start_nonterminal,
        )
        self.config = config
        self.terminal_conversions = get_terminal_conversions()
        self.nonterminal_conversions = get_nonterminal_conversions(
            config.version, config.future_imports
        )

    def convert_nonterminal(self, nonterminal: str, children: Sequence[Any]) -> Any:
        return self.nonterminal_conversions[nonterminal](self.config, children)

    def convert_terminal(self, token: Token) -> Any:
        return self.terminal_conversions[token.type.name](self.config, token)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from dataclasses import dataclass
from typing import cast, Optional, Sequence, TYPE_CHECKING, TypeVar, Union

from libcst._add_slots import add_slots
from libcst._nodes.base import CSTNode
from libcst._nodes.internal import CodegenState, visit_body_sequence, visit_sequence
from libcst._nodes.statement import (
    BaseCompoundStatement,
    get_docstring_impl,
    SimpleStatementLine,
)
from libcst._nodes.whitespace import EmptyLine
from libcst._removal_sentinel import RemovalSentinel
from libcst._visitors import CSTVisitorT

if TYPE_CHECKING:
    # This is circular, so import the type only in type checking
    from libcst._parser.types.config import PartialParserConfig


_ModuleSelfT = TypeVar("_ModuleSelfT", bound="Module")

# type alias needed for scope overlap in type definition
builtin_bytes = bytes


@add_slots
@dataclass(frozen=True)
class Module(CSTNode):
    """
    Contains some top-level information inferred from the file letting us set correct
    defaults when printing the tree about global formatting rules. All code parsed
    with :func:`parse_module` will be encapsulated in a module.
    """

    #: A list of zero or more statements that make up this module.
    body: Sequence[Union[SimpleStatementLine, BaseCompoundStatement]]

    #: Normally any whitespace/comments are assigned to the next node visited, but
    #: :class:`Module` is a special case, and comments at the top of the file tend
    #: to refer to the module itself, so we assign them to the :class:`Module`
    #: instead of the first statement in the body.
    header: Sequence[EmptyLine] = ()

    #: Any trailing whitespace/comments found after the last statement.
    footer: Sequence[EmptyLine] = ()

    #: The file's encoding format. When parsing a ``bytes`` object, this value may be
    #: inferred from the contents of the parsed source code. When parsing a ``str``,
    #: this value defaults to ``"utf-8"``.
    #:
    #: This value affects how :attr:`bytes` encodes the source code.
    encoding: str = "utf-8"

    #: The indentation of the file, expressed as a series of tabs and/or spaces. This
    #: value is inferred from the contents of the parsed source code by default.
    default_indent: str = " " * 4

    #: The newline of the file, expressed as ``\n``, ``\r\n``, or ``\r``. This value is
    #: inferred from the contents of the parsed source code by default.
    default_newline: str = "\n"

    #: Whether the module has a trailing newline or not.
    has_trailing_newline: bool = True

    def _visit_and_replace_children(self, visitor: CSTVisitorT) -> "Module":
        return Module(
            header=visit_sequence(self, "header", self.header, visitor),
            body=visit_body_sequence(self, "body", self.body, visitor),
            footer=visit_sequence(self, "footer", self.footer, visitor),
            encoding=self.encoding,
            default_indent=self.default_indent,
            default_newline=self.default_newline,
            has_trailing_newline=self.has_trailing_newline,
        )

    def visit(self: _ModuleSelfT, visitor: CSTVisitorT) -> _ModuleSelfT:
        """
        Returns the result of running a visitor over this module.

        :class:`Module` overrides the default visitor entry point to resolve metadata
        dependencies declared by 'visitor'.
        """
        result = super(Module, self).visit(visitor)
        if isinstance(result, RemovalSentinel):
            return self.with_changes(body=(), header=(), footer=())
        else:  # is a Module
            return cast(_ModuleSelfT, result)

    def _codegen_impl(self, state: CodegenState) -> None:
        for h in self.header:
            h._codegen(state)
        for stmt in self.body:
            stmt._codegen(state)
        for f in self.footer:
            f._codegen(state)
        if self.has_trailing_newline:
            if len(state.tokens) == 0:
                # There was nothing in the header, footer, or body. Just add a newline
                # to preserve the trailing newline.
                state.add_token(state.default_newline)
        else:  # has_trailing_newline is false
            state.pop_trailing_newline()

    @property
    def code(self) -> str:
        """
        The string representation of this module, respecting the inferred indentation
        and newline type.
        """
        return self.code_for_node(self)

    @property
    def bytes(self) -> builtin_bytes:
        """
        The bytes representation of this module, respecting the inferred indentation
        and newline type, using the current encoding.
        """
        return self.code.encode(self.encoding)

    def code_for_node(self, node: CSTNode) -> str:
        """
        Generates the code for the given node in the context of this module. This is a
        method of Module, not CSTNode, because we need to know the module's default
        indentation and newline formats.
        """

        state = CodegenState(
            default_indent=self.default_indent, default_newline=self.default_newline
        )
        node._codegen(state)
        return "".join(state.tokens)

    @property
    def config_for_parsing(self) -> "PartialParserConfig":
        """
        Generates a parser config appropriate for passing to a :func:`parse_expression`
        or :func:`parse_statement` call. This is useful when using either parser
        function to generate code from a string template. By using a generated parser
        config instead of the default, you can guarantee that trees generated from
        both statement and expression strings have the same inferred defaults for things
        like newlines, indents and similar::

            module = cst.parse_module("pass\\n")
            expression = cst.parse_expression("1 + 2", config=module.config_for_parsing)
        """

        from libcst._parser.types.config import PartialParserConfig

        return PartialParserConfig(
            encoding=self.encoding,
            default_indent=self.default_indent,
            default_newline=self.default_newline,
        )

    def get_docstring(self, clean: bool = True) -> Optional[str]:
        """
        Returns a :func:`inspect.cleandoc` cleaned docstring if the docstring is available, ``None`` otherwise.
        """
        return get_docstring_impl(self.body, clean)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Defines the state object used by the whitespace parser.
"""

try:
    from libcst_native import whitespace_state as mod
except ImportError:
    from libcst._parser.types import py_whitespace_state as mod

WhitespaceState = mod.WhitespaceState



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
from contextlib import contextmanager
from pathlib import Path
from typing import Generator

from libcst._types import StrPath


@contextmanager
def chdir(path: StrPath) -> Generator[Path, None, None]:
    """
    Temporarily chdir to the given path, and then return to the previous path.
    """
    try:
        path = Path(path).resolve()
        cwd = os.getcwd()
        os.chdir(path)
        yield path
    finally:
        os.chdir(cwd)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_expression
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class DictCompTest(CSTNodeTest):
    @data_provider(
        [
            # simple DictComp
            {
                "node": cst.DictComp(
                    cst.Name("k"),
                    cst.Name("v"),
                    cst.CompFor(target=cst.Name("a"), iter=cst.Name("b")),
                ),
                "code": "{k: v for a in b}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 17)),
            },
            # non-trivial keys & values in DictComp
            {
                "node": cst.DictComp(
                    cst.BinaryOperation(cst.Name("k1"), cst.Add(), cst.Name("k2")),
                    cst.BinaryOperation(cst.Name("v1"), cst.Add(), cst.Name("v2")),
                    cst.CompFor(target=cst.Name("a"), iter=cst.Name("b")),
                ),
                "code": "{k1 + k2: v1 + v2 for a in b}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 29)),
            },
            # custom whitespace around colon
            {
                "node": cst.DictComp(
                    cst.Name("k"),
                    cst.Name("v"),
                    cst.CompFor(target=cst.Name("a"), iter=cst.Name("b")),
                    whitespace_before_colon=cst.SimpleWhitespace("\t"),
                    whitespace_after_colon=cst.SimpleWhitespace("\t\t"),
                ),
                "code": "{k\t:\t\tv for a in b}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 19)),
            },
            # custom whitespace inside braces
            {
                "node": cst.DictComp(
                    cst.Name("k"),
                    cst.Name("v"),
                    cst.CompFor(target=cst.Name("a"), iter=cst.Name("b")),
                    lbrace=cst.LeftCurlyBrace(
                        whitespace_after=cst.SimpleWhitespace("\t")
                    ),
                    rbrace=cst.RightCurlyBrace(
                        whitespace_before=cst.SimpleWhitespace("\t\t")
                    ),
                ),
                "code": "{\tk: v for a in b\t\t}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 20)),
            },
            # parenthesis
            {
                "node": cst.DictComp(
                    cst.Name("k"),
                    cst.Name("v"),
                    cst.CompFor(target=cst.Name("a"), iter=cst.Name("b")),
                    lpar=[cst.LeftParen()],
                    rpar=[cst.RightParen()],
                ),
                "code": "({k: v for a in b})",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 18)),
            },
            # missing spaces around DictComp is always okay
            {
                "node": cst.DictComp(
                    cst.Name("a"),
                    cst.Name("b"),
                    cst.CompFor(
                        target=cst.Name("c"),
                        iter=cst.DictComp(
                            cst.Name("d"),
                            cst.Name("e"),
                            cst.CompFor(target=cst.Name("f"), iter=cst.Name("g")),
                        ),
                        ifs=[
                            cst.CompIf(
                                cst.Name("h"),
                                whitespace_before=cst.SimpleWhitespace(""),
                            )
                        ],
                        whitespace_after_in=cst.SimpleWhitespace(""),
                    ),
                ),
                "code": "{a: b for c in{d: e for f in g}if h}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 36)),
            },
            # no whitespace before `for` clause
            {
                "node": cst.DictComp(
                    cst.Name("k"),
                    cst.Name("v", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]),
                    cst.CompFor(
                        target=cst.Name("a"),
                        iter=cst.Name("b"),
                        whitespace_before=cst.SimpleWhitespace(""),
                    ),
                ),
                "code": "{k: (v)for a in b}",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 18)),
            },
        ]
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        [
            # unbalanced DictComp
            {
                "get_node": lambda: cst.DictComp(
                    cst.Name("k"),
                    cst.Name("v"),
                    cst.CompFor(target=cst.Name("a"), iter=cst.Name("b")),
                    lpar=[cst.LeftParen()],
                ),
                "expected_re": "left paren without right paren",
            },
            # invalid whitespace before for/async
            {
                "get_node": lambda: cst.DictComp(
                    cst.Name("k"),
                    cst.Name("v"),
                    cst.CompFor(
                        target=cst.Name("a"),
                        iter=cst.Name("b"),
                        whitespace_before=cst.SimpleWhitespace(""),
                    ),
                ),
                "expected_re": "Must have at least one space before 'for' keyword.",
            },
            {
                "get_node": lambda: cst.DictComp(
                    cst.Name("k"),
                    cst.Name("v"),
                    cst.CompFor(
                        target=cst.Name("a"),
                        iter=cst.Name("b"),
                        asynchronous=cst.Asynchronous(),
                        whitespace_before=cst.SimpleWhitespace(""),
                    ),
                ),
                "expected_re": "Must have at least one space before 'async' keyword.",
            },
        ]
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_expression
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class BooleanOperationTest(CSTNodeTest):
    @data_provider(
        (
            # Simple boolean operations
            {
                "node": cst.BooleanOperation(
                    cst.Name("foo"), cst.And(), cst.Name("bar")
                ),
                "code": "foo and bar",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.BooleanOperation(
                    cst.Name("foo"), cst.Or(), cst.Name("bar")
                ),
                "code": "foo or bar",
                "parser": parse_expression,
                "expected_position": None,
            },
            # Parenthesized boolean operation
            {
                "node": cst.BooleanOperation(
                    lpar=(cst.LeftParen(),),
                    left=cst.Name("foo"),
                    operator=cst.Or(),
                    right=cst.Name("bar"),
                    rpar=(cst.RightParen(),),
                ),
                "code": "(foo or bar)",
                "parser": parse_expression,
                "expected_position": None,
            },
            {
                "node": cst.BooleanOperation(
                    left=cst.Name(
                        "foo", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                    ),
                    operator=cst.Or(
                        whitespace_before=cst.SimpleWhitespace(""),
                        whitespace_after=cst.SimpleWhitespace(""),
                    ),
                    right=cst.Name(
                        "bar", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                    ),
                ),
                "code": "(foo)or(bar)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 12)),
            },
            # Make sure that spacing works
            {
                "node": cst.BooleanOperation(
                    lpar=(cst.LeftParen(whitespace_after=cst.SimpleWhitespace(" ")),),
                    left=cst.Name("foo"),
                    operator=cst.And(
                        whitespace_before=cst.SimpleWhitespace("  "),
                        whitespace_after=cst.SimpleWhitespace("  "),
                    ),
                    right=cst.Name("bar"),
                    rpar=(cst.RightParen(whitespace_before=cst.SimpleWhitespace(" ")),),
                ),
                "code": "( foo  and  bar )",
                "parser": parse_expression,
                "expected_position": None,
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "get_node": lambda: cst.BooleanOperation(
                    cst.Name("foo"), cst.And(), cst.Name("bar"), lpar=(cst.LeftParen(),)
                ),
                "expected_re": "left paren without right paren",
            },
            {
                "get_node": lambda: cst.BooleanOperation(
                    cst.Name("foo"),
                    cst.And(),
                    cst.Name("bar"),
                    rpar=(cst.RightParen(),),
                ),
                "expected_re": "right paren without left paren",
            },
            {
                "get_node": lambda: cst.BooleanOperation(
                    left=cst.Name("foo"),
                    operator=cst.Or(
                        whitespace_before=cst.SimpleWhitespace(""),
                        whitespace_after=cst.SimpleWhitespace(""),
                    ),
                    right=cst.Name("bar"),
                ),
                "expected_re": "at least one space around boolean operator",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from textwrap import dedent
from typing import Optional

import libcst as cst
from libcst.testing.utils import UnitTest


class DeepReplaceTest(UnitTest):
    def test_deep_replace_simple(self) -> None:
        old_code = """
            pass
        """
        new_code = """
            break
        """

        module = cst.parse_module(dedent(old_code))
        pass_stmt = cst.ensure_type(module.body[0], cst.SimpleStatementLine).body[0]
        new_module = cst.ensure_type(
            module.deep_replace(pass_stmt, cst.Break()), cst.Module
        )
        self.assertEqual(new_module.code, dedent(new_code))

    def test_deep_replace_complex(self) -> None:
        old_code = """
            def a():
                def b():
                    def c():
                        pass
        """
        new_code = """
            def a():
                def b():
                    def d(): break
        """

        module = cst.parse_module(dedent(old_code))
        outer_fun = cst.ensure_type(module.body[0], cst.FunctionDef)
        middle_fun = cst.ensure_type(
            cst.ensure_type(outer_fun.body, cst.IndentedBlock).body[0], cst.FunctionDef
        )
        inner_fun = cst.ensure_type(
            cst.ensure_type(middle_fun.body, cst.IndentedBlock).body[0], cst.FunctionDef
        )
        new_module = cst.ensure_type(
            module.deep_replace(
                inner_fun,
                cst.FunctionDef(
                    name=cst.Name("d"),
                    params=cst.Parameters(),
                    body=cst.SimpleStatementSuite(body=(cst.Break(),)),
                ),
            ),
            cst.Module,
        )
        self.assertEqual(new_module.code, dedent(new_code))

    def test_deep_replace_identity(self) -> None:
        old_code = """
            pass
        """
        new_code = """
            break
        """

        module = cst.parse_module(dedent(old_code))
        new_module = module.deep_replace(
            module,
            cst.Module(
                header=(cst.EmptyLine(),),
                body=(cst.SimpleStatementLine(body=(cst.Break(),)),),
            ),
        )
        self.assertEqual(new_module.code, dedent(new_code))

    def test_deep_remove_complex(self) -> None:
        old_code = """
            def a():
                def b():
                    def c():
                        print("Hello, world!")
        """
        new_code = """
            def a():
                def b():
                    pass
        """

        module = cst.parse_module(dedent(old_code))
        outer_fun = cst.ensure_type(module.body[0], cst.FunctionDef)
        middle_fun = cst.ensure_type(
            cst.ensure_type(outer_fun.body, cst.IndentedBlock).body[0], cst.FunctionDef
        )
        inner_fun = cst.ensure_type(
            cst.ensure_type(middle_fun.body, cst.IndentedBlock).body[0], cst.FunctionDef
        )
        new_module = cst.ensure_type(module.deep_remove(inner_fun), cst.Module)
        self.assertEqual(new_module.code, dedent(new_code))

    def test_with_deep_changes_complex(self) -> None:
        old_code = """
            def a():
                def b():
                    def c():
                        print("Hello, world!")
        """
        new_code = """
            def a():
                def b():
                    def c():
                        print("Goodbye, world!")
        """

        class NodeFinder(cst.CSTVisitor):
            # I wrote this so I wouldn't have to do a nasty multi-level
            # tree walk, but it is also a nice example of how to implement
            # a simple node find in the absence of official support.
            def __init__(self) -> None:
                super().__init__()
                self.node: Optional[cst.CSTNode] = None

            def visit_SimpleString(self, node: cst.SimpleString) -> None:
                self.node = node

        module = cst.parse_module(dedent(old_code))
        node_finder = NodeFinder()
        module.visit(node_finder)
        node = node_finder.node
        assert node is not None, "Expected to find a string node!"
        new_module = cst.ensure_type(
            module.with_deep_changes(node, value='"Goodbye, world!"'), cst.Module
        )
        self.assertEqual(new_module.code, dedent(new_code))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from os import environ

import setuptools
from setuptools_rust import Binding, RustExtension


def no_local_scheme(version: str) -> str:
    return ""


setuptools.setup(
    setup_requires=["setuptools-rust", "setuptools_scm"],
    use_scm_version={
        "write_to": "libcst/_version.py",
        **(
            {"local_scheme": no_local_scheme}
            if "LIBCST_NO_LOCAL_SCHEME" in environ
            else {}
        ),
    },
    packages=setuptools.find_packages(),
    package_data={
        "libcst": ["py.typed"],
        "libcst.tests.pyre": ["*"],
        "libcst.codemod.tests": ["*"],
    },
    test_suite="libcst",
    rust_extensions=[
        RustExtension(
            "libcst.native",
            path="native/libcst/Cargo.toml",
            binding=Binding.PyO3,
        )
    ],
    zip_safe=False,  # for mypy compatibility https://mypy.readthedocs.io/en/latest/installed_packages.html
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Callable

import libcst as cst
from libcst._nodes.tests.base import CSTNodeTest
from libcst.testing.utils import data_provider


class CommentTest(CSTNodeTest):
    @data_provider(
        (
            (cst.Comment("#"), "#"),
            (cst.Comment("#comment text"), "#comment text"),
            (cst.Comment("# comment text"), "# comment text"),
        )
    )
    def test_valid(self, node: cst.CSTNode, code: str) -> None:
        self.validate_node(node, code)

    @data_provider(
        (
            (lambda: cst.Comment(" bad input"), "non-comment"),
            (lambda: cst.Comment("# newline shouldn't be here\n"), "non-comment"),
            (lambda: cst.Comment(" # Leading space is wrong"), "non-comment"),
        )
    )
    def test_invalid(
        self, get_node: Callable[[], cst.CSTNode], expected_re: str
    ) -> None:
        self.assert_invalid(get_node, expected_re)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
# pyre-unsafe

from typing import Any

from libcst._nodes.expression import SimpleString
from libcst._parser.types.config import ParserConfig
from libcst._parser.types.partials import WithLeadingWhitespace
from libcst._parser.types.token import Token
from libcst._parser.whitespace_parser import (
    parse_empty_lines,
    parse_trailing_whitespace,
)


def convert_NAME(config: ParserConfig, token: Token) -> Any:
    return token


def convert_NUMBER(config: ParserConfig, token: Token) -> Any:
    return token


def convert_STRING(config: ParserConfig, token: Token) -> Any:
    return WithLeadingWhitespace(SimpleString(token.string), token.whitespace_before)


def convert_OP(config: ParserConfig, token: Token) -> Any:
    return token


def convert_NEWLINE(config: ParserConfig, token: Token) -> Any:
    # A NEWLINE token is only emitted for semantic newlines, which means that this
    # corresponds to a TrailingWhitespace, since that's the only semantic
    # newline-containing node.

    # N.B. Because this token is whitespace, and because the whitespace parser doesn't
    # try to prevent overflows, `token.whitespace_before` will end up overflowing into
    # the value of this newline token, so `parse_trailing_whitespace` will include
    # token.string's value. This is expected and desired behavior.
    return parse_trailing_whitespace(config, token.whitespace_before)


def convert_INDENT(config: ParserConfig, token: Token) -> Any:
    return token


def convert_DEDENT(config: ParserConfig, token: Token) -> Any:
    return token


def convert_ENDMARKER(config: ParserConfig, token: Token) -> Any:
    # Parse any and all empty lines with an indent similar to the header. That is,
    # indent of nothing and including all indents. In some cases, like when the
    # footer parser follows an indented suite, the state's indent can be wrong
    # due to the fact that it is shared with the _DEDENT node. We know that if
    # we're parsing the end of a file, we will have no indent.
    return parse_empty_lines(
        config, token.whitespace_before, override_absolute_indent=""
    )


def convert_FSTRING_START(config: ParserConfig, token: Token) -> Any:
    return token


def convert_FSTRING_END(config: ParserConfig, token: Token) -> Any:
    return token


def convert_FSTRING_STRING(config: ParserConfig, token: Token) -> Any:
    return token


def convert_ASYNC(config: ParserConfig, token: Token) -> Any:
    return token


def convert_AWAIT(config: ParserConfig, token: Token) -> Any:
    return token



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any

import libcst as cst
from libcst import parse_statement, PartialParserConfig
from libcst._nodes.tests.base import CSTNodeTest, DummyIndentedBlock
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class ForTest(CSTNodeTest):
    @data_provider(
        (
            # Simple for block
            {
                "node": cst.For(
                    cst.Name("target"),
                    cst.Call(cst.Name("iter")),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                ),
                "code": "for target in iter(): pass\n",
                "parser": parse_statement,
            },
            # Simple async for block
            {
                "node": cst.For(
                    cst.Name("target"),
                    cst.Call(cst.Name("iter")),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    asynchronous=cst.Asynchronous(),
                ),
                "code": "async for target in iter(): pass\n",
                "parser": lambda code: parse_statement(
                    code, config=PartialParserConfig(python_version="3.7")
                ),
            },
            # Python 3.6 async for block
            {
                "node": cst.FunctionDef(
                    cst.Name("foo"),
                    cst.Parameters(),
                    cst.IndentedBlock(
                        (
                            cst.For(
                                cst.Name("target"),
                                cst.Call(cst.Name("iter")),
                                cst.SimpleStatementSuite((cst.Pass(),)),
                                asynchronous=cst.Asynchronous(),
                            ),
                        )
                    ),
                    asynchronous=cst.Asynchronous(),
                ),
                "code": "async def foo():\n    async for target in iter(): pass\n",
                "parser": lambda code: parse_statement(
                    code, config=PartialParserConfig(python_version="3.6")
                ),
            },
            # For block with else
            {
                "node": cst.For(
                    cst.Name("target"),
                    cst.Call(cst.Name("iter")),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    cst.Else(cst.SimpleStatementSuite((cst.Pass(),))),
                ),
                "code": "for target in iter(): pass\nelse: pass\n",
                "parser": parse_statement,
            },
            # indentation
            {
                "node": DummyIndentedBlock(
                    "    ",
                    cst.For(
                        cst.Name("target"),
                        cst.Call(cst.Name("iter")),
                        cst.SimpleStatementSuite((cst.Pass(),)),
                    ),
                ),
                "code": "    for target in iter(): pass\n",
                "parser": None,
            },
            # for an indented body
            {
                "node": DummyIndentedBlock(
                    "    ",
                    cst.For(
                        cst.Name("target"),
                        cst.Call(cst.Name("iter")),
                        cst.IndentedBlock((cst.SimpleStatementLine((cst.Pass(),)),)),
                    ),
                ),
                "code": "    for target in iter():\n        pass\n",
                "parser": None,
                "expected_position": CodeRange((1, 4), (2, 12)),
            },
            # leading_lines
            {
                "node": cst.For(
                    cst.Name("target"),
                    cst.Call(cst.Name("iter")),
                    cst.IndentedBlock((cst.SimpleStatementLine((cst.Pass(),)),)),
                    cst.Else(
                        cst.IndentedBlock((cst.SimpleStatementLine((cst.Pass(),)),)),
                        leading_lines=(
                            cst.EmptyLine(comment=cst.Comment("# else comment")),
                        ),
                    ),
                    leading_lines=(
                        cst.EmptyLine(comment=cst.Comment("# leading comment")),
                    ),
                ),
                "code": "# leading comment\nfor target in iter():\n    pass\n# else comment\nelse:\n    pass\n",
                "parser": None,
                "expected_position": CodeRange((2, 0), (6, 8)),
            },
            # Weird spacing rules
            {
                "node": cst.For(
                    cst.Name(
                        "target", lpar=(cst.LeftParen(),), rpar=(cst.RightParen(),)
                    ),
                    cst.Call(
                        cst.Name("iter"),
                        lpar=(cst.LeftParen(),),
                        rpar=(cst.RightParen(),),
                    ),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    whitespace_after_for=cst.SimpleWhitespace(""),
                    whitespace_before_in=cst.SimpleWhitespace(""),
                    whitespace_after_in=cst.SimpleWhitespace(""),
                ),
                "code": "for(target)in(iter()): pass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (1, 27)),
            },
            # Whitespace
            {
                "node": cst.For(
                    cst.Name("target"),
                    cst.Call(cst.Name("iter")),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    whitespace_after_for=cst.SimpleWhitespace("  "),
                    whitespace_before_in=cst.SimpleWhitespace("  "),
                    whitespace_after_in=cst.SimpleWhitespace("  "),
                    whitespace_before_colon=cst.SimpleWhitespace("  "),
                ),
                "code": "for  target  in  iter()  : pass\n",
                "parser": parse_statement,
                "expected_position": CodeRange((1, 0), (1, 31)),
            },
        )
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            {
                "get_node": lambda: cst.For(
                    cst.Name("target"),
                    cst.Call(cst.Name("iter")),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    whitespace_after_for=cst.SimpleWhitespace(""),
                ),
                "expected_re": "Must have at least one space after 'for' keyword",
            },
            {
                "get_node": lambda: cst.For(
                    cst.Name("target"),
                    cst.Call(cst.Name("iter")),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    whitespace_before_in=cst.SimpleWhitespace(""),
                ),
                "expected_re": "Must have at least one space before 'in' keyword",
            },
            {
                "get_node": lambda: cst.For(
                    cst.Name("target"),
                    cst.Call(cst.Name("iter")),
                    cst.SimpleStatementSuite((cst.Pass(),)),
                    whitespace_after_in=cst.SimpleWhitespace(""),
                ),
                "expected_re": "Must have at least one space after 'in' keyword",
            },
        )
    )
    def test_invalid(self, **kwargs: Any) -> None:
        self.assert_invalid(**kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from textwrap import dedent

import libcst as cst
from libcst import parse_module
from libcst._nodes.deep_equals import deep_equals
from libcst.testing.utils import data_provider, UnitTest


class FooterBehaviorTest(UnitTest):
    @data_provider(
        {
            # Literally the most basic example
            "simple_module": {
                "code": "",
                "expected_module": cst.Module(body=(), has_trailing_newline=False),
            },
            # A module with a header comment
            "header_only_module": {
                "code": "# This is a header comment\n",
                "expected_module": cst.Module(
                    header=[
                        cst.EmptyLine(
                            comment=cst.Comment(value="# This is a header comment")
                        )
                    ],
                    body=[],
                ),
            },
            # A module with a header and footer
            "simple_header_footer_module": {
                "code": "# This is a header comment\npass\n# This is a footer comment\n",
                "expected_module": cst.Module(
                    header=[
                        cst.EmptyLine(
                            comment=cst.Comment(value="# This is a header comment")
                        )
                    ],
                    body=[cst.SimpleStatementLine([cst.Pass()])],
                    footer=[
                        cst.EmptyLine(
                            comment=cst.Comment(value="# This is a footer comment")
                        )
                    ],
                ),
            },
            # A module which should have a footer comment taken from the
            # if statement's indented block.
            "simple_reparented_footer_module": {
                "code": "# This is a header comment\nif True:\n    pass\n# This is a footer comment\n",
                "expected_module": cst.Module(
                    header=[
                        cst.EmptyLine(
                            comment=cst.Comment(value="# This is a header comment")
                        )
                    ],
                    body=[
                        cst.If(
                            test=cst.Name(value="True"),
                            body=cst.IndentedBlock(
                                header=cst.TrailingWhitespace(),
                                body=[
                                    cst.SimpleStatementLine(
                                        body=[cst.Pass()],
                                        trailing_whitespace=cst.TrailingWhitespace(),
                                    )
                                ],
                            ),
                        )
                    ],
                    footer=[
                        cst.EmptyLine(
                            comment=cst.Comment(value="# This is a footer comment")
                        )
                    ],
                ),
            },
            # Verifying that we properly parse and spread out footer comments to the
            # relative indents they go with.
            "complex_reparented_footer_module": {
                "code": (
                    "# This is a header comment\nif True:\n    if True:\n        pass"
                    + "\n        # This is an inner indented block comment\n    # This "
                    + "is an outer indented block comment\n# This is a footer comment\n"
                ),
                "expected_module": cst.Module(
                    body=[
                        cst.If(
                            test=cst.Name(value="True"),
                            body=cst.IndentedBlock(
                                body=[
                                    cst.If(
                                        test=cst.Name(value="True"),
                                        body=cst.IndentedBlock(
                                            body=[
                                                cst.SimpleStatementLine(
                                                    body=[cst.Pass()]
                                                )
                                            ],
                                            footer=[
                                                cst.EmptyLine(
                                                    comment=cst.Comment(
                                                        value="# This is an inner indented block comment"
                                                    )
                                                )
                                            ],
                                        ),
                                    )
                                ],
                                footer=[
                                    cst.EmptyLine(
                                        comment=cst.Comment(
                                            value="# This is an outer indented block comment"
                                        )
                                    )
                                ],
                            ),
                        )
                    ],
                    header=[
                        cst.EmptyLine(
                            comment=cst.Comment(value="# This is a header comment")
                        )
                    ],
                    footer=[
                        cst.EmptyLine(
                            comment=cst.Comment(value="# This is a footer comment")
                        )
                    ],
                ),
            },
            # Verify that comments belonging to statements are still owned even
            # after an indented block.
            "statement_comment_reparent": {
                "code": "if foo:\n    return\n# comment\nx = 7\n",
                "expected_module": cst.Module(
                    body=[
                        cst.If(
                            test=cst.Name(value="foo"),
                            body=cst.IndentedBlock(
                                body=[
                                    cst.SimpleStatementLine(
                                        body=[
                                            cst.Return(
                                                whitespace_after_return=cst.SimpleWhitespace(
                                                    value=""
                                                )
                                            )
                                        ]
                                    )
                                ]
                            ),
                        ),
                        cst.SimpleStatementLine(
                            body=[
                                cst.Assign(
                                    targets=[
                                        cst.AssignTarget(target=cst.Name(value="x"))
                                    ],
                                    value=cst.Integer(value="7"),
                                )
                            ],
                            leading_lines=[
                                cst.EmptyLine(comment=cst.Comment(value="# comment"))
                            ],
                        ),
                    ]
                ),
            },
            # Verify that even if there are completely empty lines, we give all lines
            # up to and including the last line that's indented correctly. That way
            # comments that line up with indented block's indentation level aren't
            # parented to the next line just because there's a blank line or two
            # between them.
            "statement_comment_with_empty_lines": {
                "code": (
                    "def foo():\n    if True:\n        pass\n\n        # Empty "
                    + "line before me\n\n    else:\n        pass\n"
                ),
                "expected_module": cst.Module(
                    body=[
                        cst.FunctionDef(
                            name=cst.Name(value="foo"),
                            params=cst.Parameters(),
                            body=cst.IndentedBlock(
                                body=[
                                    cst.If(
                                        test=cst.Name(value="True"),
                                        body=cst.IndentedBlock(
                                            body=[
                                                cst.SimpleStatementLine(
                                                    body=[cst.Pass()]
                                                )
                                            ],
                                            footer=[
                                                cst.EmptyLine(indent=False),
                                                cst.EmptyLine(
                                                    comment=cst.Comment(
                                                        value="# Empty line before me"
                                                    )
                                                ),
                                            ],
                                        ),
                                        orelse=cst.Else(
                                            body=cst.IndentedBlock(
                                                body=[
                                                    cst.SimpleStatementLine(
                                                        body=[cst.Pass()]
                                                    )
                                                ]
                                            ),
                                            leading_lines=[cst.EmptyLine(indent=False)],
                                        ),
                                    )
                                ]
                            ),
                        )
                    ]
                ),
            },
        }
    )
    def test_parsers(self, code: str, expected_module: cst.CSTNode) -> None:
        parsed_module = parse_module(dedent(code))
        self.assertTrue(
            deep_equals(parsed_module, expected_module),
            msg=f"\n{parsed_module!r}\nis not deeply equal to \n{expected_module!r}",
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict
from libcst._parser.parso.utils import PythonVersionInfo
from libcst._parser.types.config import _pick_compatible_python_version
from libcst.testing.utils import UnitTest


class ConfigTest(UnitTest):
    def test_pick_compatible(self) -> None:
        self.assertEqual(
            PythonVersionInfo(3, 1), _pick_compatible_python_version("3.2")
        )
        self.assertEqual(
            PythonVersionInfo(3, 1), _pick_compatible_python_version("3.1")
        )
        self.assertEqual(
            PythonVersionInfo(3, 8), _pick_compatible_python_version("3.9")
        )
        self.assertEqual(
            PythonVersionInfo(3, 8), _pick_compatible_python_version("3.10")
        )
        self.assertEqual(
            PythonVersionInfo(3, 8), _pick_compatible_python_version("4.0")
        )
        with self.assertRaisesRegex(
            ValueError,
            (
                r"No version found older than 1\.0 \(PythonVersionInfo\("
                + r"major=1, minor=0\)\) while running on"
            ),
        ):
            _pick_compatible_python_version("1.0")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from contextlib import contextmanager
from dataclasses import dataclass, field
from typing import Callable, Iterator, List, Optional

from libcst import CSTNode, Module
from libcst._nodes.internal import CodegenState
from libcst.metadata.base_provider import BaseMetadataProvider


@dataclass(frozen=True)
class CodeSpan:
    """
    Represents the position of a piece of code by its starting position and length.

    Note: This class does not specify the unit of distance - it can be bytes,
    Unicode characters, or something else entirely.
    """

    #: Offset of the code from the beginning of the file. Can be 0.
    start: int
    #: Length of the span
    length: int


@dataclass(frozen=False)
class SpanProvidingCodegenState(CodegenState):
    provider: BaseMetadataProvider[CodeSpan]
    get_length: Optional[Callable[[str], int]] = None
    position: int = 0
    _stack: List[int] = field(default_factory=list)

    def add_indent_tokens(self) -> None:
        super().add_indent_tokens()
        for token in self.indent_tokens:
            self._update_position(token)

    def add_token(self, value: str) -> None:
        super().add_token(value)
        self._update_position(value)

    def _update_position(self, value: str) -> None:
        get_length = self.get_length or len
        self.position += get_length(value)

    def before_codegen(self, node: CSTNode) -> None:
        self._stack.append(self.position)

    def after_codegen(self, node: CSTNode) -> None:
        start = self._stack.pop()

        if node not in self.provider._computed:
            end = self.position
            self.provider._computed[node] = CodeSpan(start, length=end - start)

    @contextmanager
    def record_syntactic_position(
        self,
        node: CSTNode,
        *,
        start_node: Optional[CSTNode] = None,
        end_node: Optional[CSTNode] = None,
    ) -> Iterator[None]:
        start = self.position
        try:
            yield
        finally:
            end = self.position
            start = (
                self.provider._computed[start_node].start
                if start_node is not None
                else start
            )
            if end_node is not None:
                end_span = self.provider._computed[end_node]
                length = (end_span.start + end_span.length) - start
            else:
                length = end - start
            self.provider._computed[node] = CodeSpan(start, length=length)


def byte_length_in_utf8(value: str) -> int:
    return len(value.encode("utf8"))


class ByteSpanPositionProvider(BaseMetadataProvider[CodeSpan]):
    """
    Generates offset and length metadata for nodes' positions.

    For each :class:`CSTNode` this provider generates a :class:`CodeSpan` that
    contains the byte-offset of the node from the start of the file, and its
    length (also in bytes). The whitespace owned by the node is not included in
    this length.

    Note: offset and length measure bytes, not characters (which is significant for
    example in the case of Unicode characters encoded in more than one byte)
    """

    def _gen_impl(self, module: Module) -> None:
        state = SpanProvidingCodegenState(
            default_indent=module.default_indent,
            default_newline=module.default_newline,
            provider=self,
            get_length=byte_length_in_utf8,
        )
        module._codegen(state)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

class WhitespaceState:
    def __new__(
        cls, line: int, column: int, absolute_indent: str, is_parenthesized: bool
    ) -> WhitespaceState: ...
    line: int  # one-indexed (to match parso's behavior)
    column: int  # zero-indexed (to match parso's behavior)
    # What to look for when executing `_parse_indent`.
    absolute_indent: str
    is_parenthesized: bool



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


import sys
from textwrap import dedent
from typing import cast, Mapping, Sequence, Tuple
from unittest import mock

import libcst as cst
from libcst import ensure_type
from libcst._parser.entrypoints import is_native
from libcst.metadata import MetadataWrapper
from libcst.metadata.scope_provider import (
    _gen_dotted_names,
    AnnotationScope,
    Assignment,
    BuiltinAssignment,
    BuiltinScope,
    ClassScope,
    ComprehensionScope,
    FunctionScope,
    GlobalScope,
    ImportAssignment,
    LocalScope,
    QualifiedName,
    QualifiedNameSource,
    Scope,
    ScopeProvider,
)
from libcst.testing.utils import data_provider, UnitTest


class DependentVisitor(cst.CSTVisitor):
    METADATA_DEPENDENCIES = (ScopeProvider,)


def get_scope_metadata_provider(
    module_str: str,
) -> Tuple[cst.Module, Mapping[cst.CSTNode, Scope]]:
    wrapper = MetadataWrapper(cst.parse_module(dedent(module_str)))
    return (
        wrapper.module,
        cast(
            Mapping[cst.CSTNode, Scope], wrapper.resolve(ScopeProvider)
        ),  # we're sure every node has an associated scope
    )


class ScopeProviderTest(UnitTest):
    def test_not_in_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            pass
            """
        )
        global_scope = scopes[m]
        self.assertEqual(global_scope["not_in_scope"], set())

    def test_accesses(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            foo = 'toplevel'
            fn1(foo)
            fn2(foo)
            def fn_def():
                foo = 'shadow'
                fn3(foo)
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        global_foo_assignments = list(scope_of_module["foo"])
        self.assertEqual(len(global_foo_assignments), 1)
        foo_assignment = global_foo_assignments[0]
        self.assertEqual(len(foo_assignment.references), 2)
        fn1_call_arg = ensure_type(
            ensure_type(
                ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.Call,
        ).args[0]

        fn2_call_arg = ensure_type(
            ensure_type(
                ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.Call,
        ).args[0]
        self.assertEqual(
            {access.node for access in foo_assignment.references},
            {fn1_call_arg.value, fn2_call_arg.value},
        )
        func_body = ensure_type(m.body[3], cst.FunctionDef).body
        func_foo_statement = func_body.body[0]
        scope_of_func_statement = scopes[func_foo_statement]
        self.assertIsInstance(scope_of_func_statement, FunctionScope)
        func_foo_assignments = scope_of_func_statement["foo"]
        self.assertEqual(len(func_foo_assignments), 1)
        foo_assignment = list(func_foo_assignments)[0]
        self.assertEqual(len(foo_assignment.references), 1)
        fn3_call_arg = ensure_type(
            ensure_type(
                ensure_type(func_body.body[1], cst.SimpleStatementLine).body[0],
                cst.Expr,
            ).value,
            cst.Call,
        ).args[0]
        self.assertEqual(
            {access.node for access in foo_assignment.references}, {fn3_call_arg.value}
        )

        wrapper = MetadataWrapper(cst.parse_module("from a import b\n"))
        wrapper.visit(DependentVisitor())

        wrapper = MetadataWrapper(cst.parse_module("def a():\n    from b import c\n\n"))
        wrapper.visit(DependentVisitor())

    def test_fstring_accesses(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            from a import b
            f"{b}" "hello"
            """
        )
        global_scope = scopes[m]
        self.assertIsInstance(global_scope, GlobalScope)
        global_accesses = list(global_scope.accesses)
        self.assertEqual(len(global_accesses), 1)
        import_node = ensure_type(
            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.ImportFrom
        )
        b_referent = list(global_accesses[0].referents)[0]
        self.assertIsInstance(b_referent, Assignment)
        if isinstance(b_referent, Assignment):  # for the typechecker's eyes
            self.assertEqual(b_referent.node, import_node)

    @data_provider((("any",), ("True",), ("Exception",), ("__name__",)))
    def test_builtins(self, builtin: str) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            def fn():
                pass
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        self.assertEqual(len(scope_of_module[builtin]), 1)
        self.assertEqual(len(scope_of_module["something_not_a_builtin"]), 0)

        scope_of_builtin = scope_of_module.parent
        self.assertIsInstance(scope_of_builtin, BuiltinScope)
        self.assertEqual(len(scope_of_builtin[builtin]), 1)
        self.assertEqual(len(scope_of_builtin["something_not_a_builtin"]), 0)

        func_body = ensure_type(m.body[0], cst.FunctionDef).body
        func_pass_statement = func_body.body[0]
        scope_of_func_statement = scopes[func_pass_statement]
        self.assertIsInstance(scope_of_func_statement, FunctionScope)
        self.assertEqual(len(scope_of_func_statement[builtin]), 1)
        self.assertEqual(len(scope_of_func_statement["something_not_a_builtin"]), 0)

    def test_import(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            import foo.bar
            import fizz.buzz as fizzbuzz
            import a.b.c
            import d.e.f as g
            """
        )
        scope_of_module = scopes[m]

        import_0 = cst.ensure_type(
            cst.ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.Import
        )
        self.assertEqual(scopes[import_0], scope_of_module)
        import_aliases = import_0.names
        if not isinstance(import_aliases, cst.ImportStar):
            for alias in import_aliases:
                self.assertEqual(scopes[alias], scope_of_module)

        for idx, in_scopes in enumerate(
            [
                ["foo", "foo.bar"],
                ["fizzbuzz"],
                ["a", "a.b", "a.b.c"],
                ["g"],
            ]
        ):
            for in_scope in in_scopes:
                self.assertEqual(
                    len(scope_of_module[in_scope]), 1, f"{in_scope} should be in scope."
                )

                assignment = cast(ImportAssignment, list(scope_of_module[in_scope])[0])
                self.assertEqual(
                    assignment.name,
                    in_scope,
                    f"ImportAssignment name {assignment.name} should equal to {in_scope}.",
                )
                import_node = ensure_type(m.body[idx], cst.SimpleStatementLine).body[0]
                self.assertEqual(
                    assignment.node,
                    import_node,
                    f"The node of ImportAssignment {assignment.node} should equal to {import_node}",
                )
                self.assertTrue(isinstance(import_node, (cst.Import, cst.ImportFrom)))

                names = import_node.names

                self.assertFalse(isinstance(names, cst.ImportStar))

                alias = names[0]
                as_name = alias.asname.name if alias.asname else alias.name
                self.assertEqual(
                    assignment.as_name,
                    as_name,
                    f"The alias name of ImportAssignment {assignment.as_name} should equal to {as_name}",
                )

    def test_dotted_import_access(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            import a.b.c, x.y
            a.b.c(x.z)
            """
        )
        scope_of_module = scopes[m]
        first_statement = ensure_type(m.body[1], cst.SimpleStatementLine)
        call = ensure_type(
            ensure_type(first_statement.body[0], cst.Expr).value, cst.Call
        )
        self.assertTrue("a.b.c" in scope_of_module)
        self.assertTrue("a" in scope_of_module)
        self.assertEqual(scope_of_module.accesses["a"], set())

        a_b_c_assignment = cast(ImportAssignment, list(scope_of_module["a.b.c"])[0])
        a_b_c_access = list(a_b_c_assignment.references)[0]
        self.assertEqual(scope_of_module.accesses["a.b.c"], {a_b_c_access})
        self.assertEqual(a_b_c_access.node, call.func)

        x_assignment = cast(Assignment, list(scope_of_module["x"])[0])
        x_access = list(x_assignment.references)[0]
        self.assertEqual(scope_of_module.accesses["x"], {x_access})
        self.assertEqual(
            x_access.node, ensure_type(call.args[0].value, cst.Attribute).value
        )

        self.assertTrue("x.y" in scope_of_module)
        self.assertEqual(list(scope_of_module["x.y"])[0].references, set())
        self.assertEqual(scope_of_module.accesses["x.y"], set())

    def test_dotted_import_access_reference_by_node(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            import a.b.c
            a.b.c()
            """
        )
        scope_of_module = scopes[m]
        first_statement = ensure_type(m.body[1], cst.SimpleStatementLine)
        call = ensure_type(
            ensure_type(first_statement.body[0], cst.Expr).value, cst.Call
        )

        a_b_c_assignment = cast(ImportAssignment, list(scope_of_module["a.b.c"])[0])
        a_b_c_access = list(a_b_c_assignment.references)[0]
        self.assertEqual(scope_of_module.accesses[call], {a_b_c_access})
        self.assertEqual(a_b_c_access.node, call.func)

    def test_decorator_access_reference_by_node(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            import decorator

            @decorator
            def f():
                pass
            """
        )
        scope_of_module = scopes[m]
        function_def = ensure_type(m.body[1], cst.FunctionDef)
        decorator = function_def.decorators[0]
        self.assertTrue("decorator" in scope_of_module)

        decorator_assignment = cast(
            ImportAssignment, list(scope_of_module["decorator"])[0]
        )
        decorator_access = list(decorator_assignment.references)[0]
        self.assertEqual(scope_of_module.accesses[decorator], {decorator_access})

    def test_dotted_import_with_call_access(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            import os.path
            os.path.join("A", "B").lower()
            """
        )
        scope_of_module = scopes[m]
        first_statement = ensure_type(m.body[1], cst.SimpleStatementLine)
        attr = ensure_type(
            ensure_type(
                ensure_type(
                    ensure_type(
                        ensure_type(first_statement.body[0], cst.Expr).value, cst.Call
                    ).func,
                    cst.Attribute,
                ).value,
                cst.Call,
            ).func,
            cst.Attribute,
        ).value
        self.assertTrue("os.path" in scope_of_module)
        self.assertTrue("os" in scope_of_module)

        os_path_join_assignment = cast(
            ImportAssignment, list(scope_of_module["os.path"])[0]
        )
        os_path_join_assignment_references = list(os_path_join_assignment.references)
        self.assertNotEqual(len(os_path_join_assignment_references), 0)
        os_path_join_access = os_path_join_assignment_references[0]
        self.assertEqual(scope_of_module.accesses["os"], set())
        self.assertEqual(scope_of_module.accesses["os.path"], {os_path_join_access})
        self.assertEqual(scope_of_module.accesses["os.path.join"], set())
        self.assertEqual(os_path_join_access.node, attr)

    def test_import_from(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            from foo.bar import a, b as b_renamed
            from . import c
            from .foo import d
            """
        )
        scope_of_module = scopes[m]

        import_from = cst.ensure_type(
            cst.ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.ImportFrom
        )
        self.assertEqual(scopes[import_from], scope_of_module)
        import_aliases = import_from.names
        if not isinstance(import_aliases, cst.ImportStar):
            for alias in import_aliases:
                self.assertEqual(scopes[alias], scope_of_module)

        for idx, in_scope, imported_object_idx in [
            (0, "a", 0),
            (0, "b_renamed", 1),
            (1, "c", 0),
            (2, "d", 0),
        ]:
            self.assertEqual(
                len(scope_of_module[in_scope]), 1, f"{in_scope} should be in scope."
            )
            import_assignment = cast(
                ImportAssignment, list(scope_of_module[in_scope])[0]
            )
            self.assertEqual(
                import_assignment.name,
                in_scope,
                f"The name of ImportAssignment {import_assignment.name} should equal to {in_scope}.",
            )
            import_node = ensure_type(m.body[idx], cst.SimpleStatementLine).body[0]
            self.assertEqual(
                import_assignment.node,
                import_node,
                f"The node of ImportAssignment {import_assignment.node} should equal to {import_node}",
            )

            self.assertTrue(isinstance(import_node, (cst.Import, cst.ImportFrom)))

            names = import_node.names

            self.assertFalse(isinstance(names, cst.ImportStar))

            alias = names[imported_object_idx]
            as_name = alias.asname.name if alias.asname else alias.name
            self.assertEqual(
                import_assignment.as_name,
                as_name,
                f"The alias name of ImportAssignment {import_assignment.as_name} should equal to {as_name}",
            )

        for not_in_scope in ["foo", "bar", "foo.bar", "b"]:
            self.assertEqual(
                len(scope_of_module[not_in_scope]),
                0,
                f"{not_in_scope} should not be in scope.",
            )

    def test_function_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            global_var = None
            def foo(arg, **kwargs):
                local_var = 5
            """
        )
        scope_of_module = scopes[m]
        func_def = ensure_type(m.body[1], cst.FunctionDef)
        self.assertEqual(scopes[func_def], scopes[func_def.name])
        func_body_statement = func_def.body.body[0]
        scope_of_func = scopes[func_body_statement]
        self.assertIsInstance(scope_of_func, FunctionScope)
        self.assertTrue("global_var" in scope_of_module)
        self.assertTrue("global_var" in scope_of_func)
        self.assertTrue("arg" not in scope_of_module)
        self.assertTrue("arg" in scope_of_func)
        self.assertTrue("kwargs" not in scope_of_module)
        self.assertTrue("kwargs" in scope_of_func)
        self.assertTrue("local_var" not in scope_of_module)
        self.assertTrue("local_var" in scope_of_func)

    def test_class_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            global_var = None
            @cls_attr
            class Cls(cls_attr, kwarg=cls_attr):
                cls_attr = 5
                def f():
                    pass
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        cls_assignments = list(scope_of_module["Cls"])
        self.assertEqual(len(cls_assignments), 1)
        cls_assignment = cast(Assignment, cls_assignments[0])
        cls_def = ensure_type(m.body[1], cst.ClassDef)
        self.assertEqual(cls_assignment.node, cls_def)
        self.assertEqual(scopes[cls_def], scopes[cls_def.name])
        cls_body = cls_def.body
        cls_body_statement = cls_body.body[0]
        scope_of_class = scopes[cls_body_statement]
        self.assertIsInstance(scope_of_class, ClassScope)
        func_body = ensure_type(cls_body.body[1], cst.FunctionDef).body
        func_body_statement = func_body.body[0]
        scope_of_func = scopes[func_body_statement]
        self.assertIsInstance(scope_of_func, FunctionScope)
        self.assertTrue("global_var" in scope_of_module)
        self.assertTrue("global_var" in scope_of_class)
        self.assertTrue("global_var" in scope_of_func)
        self.assertTrue("Cls" in scope_of_module)
        self.assertTrue("Cls" in scope_of_class)
        self.assertTrue("Cls" in scope_of_func)
        self.assertTrue("cls_attr" not in scope_of_module)
        self.assertTrue("cls_attr" in scope_of_class)
        self.assertTrue("cls_attr" not in scope_of_func)

    def test_comprehension_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            iterator = None
            condition = None
            [elt for target in iterator if condition]
            {elt for target in iterator if condition}
            {elt: target for target in iterator if condition}
            (elt for target in iterator if condition)
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)

        list_comp = ensure_type(
            ensure_type(
                ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.ListComp,
        )
        scope_of_list_comp = scopes[list_comp.elt]
        self.assertIsInstance(scope_of_list_comp, ComprehensionScope)

        set_comp = ensure_type(
            ensure_type(
                ensure_type(m.body[3], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.SetComp,
        )
        scope_of_set_comp = scopes[set_comp.elt]
        self.assertIsInstance(scope_of_set_comp, ComprehensionScope)

        dict_comp = ensure_type(
            ensure_type(
                ensure_type(m.body[4], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.DictComp,
        )
        scope_of_dict_comp = scopes[dict_comp.key]
        self.assertIsInstance(scope_of_dict_comp, ComprehensionScope)

        generator_expr = ensure_type(
            ensure_type(
                ensure_type(m.body[5], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.GeneratorExp,
        )
        scope_of_generator_expr = scopes[generator_expr.elt]
        self.assertIsInstance(scope_of_generator_expr, ComprehensionScope)

        self.assertTrue("iterator" in scope_of_module)
        self.assertTrue("iterator" in scope_of_list_comp)
        self.assertTrue("iterator" in scope_of_set_comp)
        self.assertTrue("iterator" in scope_of_dict_comp)
        self.assertTrue("iterator" in scope_of_generator_expr)

        self.assertTrue("condition" in scope_of_module)
        self.assertTrue("condition" in scope_of_list_comp)
        self.assertTrue("condition" in scope_of_set_comp)
        self.assertTrue("condition" in scope_of_dict_comp)
        self.assertTrue("condition" in scope_of_generator_expr)

        self.assertTrue("elt" not in scope_of_module)
        self.assertTrue("elt" not in scope_of_list_comp)
        self.assertTrue("elt" not in scope_of_set_comp)
        self.assertTrue("elt" not in scope_of_dict_comp)
        self.assertTrue("elt" not in scope_of_generator_expr)

        self.assertTrue("target" not in scope_of_module)
        self.assertTrue("target" in scope_of_list_comp)
        self.assertTrue("target" in scope_of_set_comp)
        self.assertTrue("target" in scope_of_dict_comp)
        self.assertTrue("target" in scope_of_generator_expr)

    def test_nested_comprehension_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            [y for x in iterator for y in x]
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)

        list_comp = ensure_type(
            ensure_type(
                ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.ListComp,
        )
        scope_of_list_comp = scopes[list_comp.elt]
        self.assertIsInstance(scope_of_list_comp, ComprehensionScope)

        self.assertIs(scopes[list_comp], scope_of_module)
        self.assertIs(scopes[list_comp.elt], scope_of_list_comp)

        self.assertIs(scopes[list_comp.for_in], scope_of_module)
        self.assertIs(scopes[list_comp.for_in.iter], scope_of_module)
        self.assertIs(scopes[list_comp.for_in.target], scope_of_list_comp)

        inner_for_in = ensure_type(list_comp.for_in.inner_for_in, cst.CompFor)
        self.assertIs(scopes[inner_for_in], scope_of_list_comp)
        self.assertIs(scopes[inner_for_in.iter], scope_of_list_comp)
        self.assertIs(scopes[inner_for_in.target], scope_of_list_comp)

    def test_global_scope_overwrites(self) -> None:
        codes = (
            """
            class Cls:
                def f():
                    global var
                    var = ...
            """,
            """
            class Cls:
                def f():
                    global var
                    import f as var
            """,
        )
        for code in codes:
            m, scopes = get_scope_metadata_provider(code)
            scope_of_module = scopes[m]
            self.assertIsInstance(scope_of_module, GlobalScope)
            self.assertTrue("var" in scope_of_module)

            cls = ensure_type(m.body[0], cst.ClassDef)
            scope_of_cls = scopes[cls.body.body[0]]
            self.assertIsInstance(scope_of_cls, ClassScope)
            self.assertTrue("var" in scope_of_cls)

            f = ensure_type(cls.body.body[0], cst.FunctionDef)
            scope_of_f = scopes[f.body.body[0]]
            self.assertIsInstance(scope_of_f, FunctionScope)
            self.assertTrue("var" in scope_of_f)
            self.assertEqual(scope_of_f["var"], scope_of_module["var"])

    def test_nonlocal_scope_overwrites(self) -> None:
        codes = (
            """
            def outer_f():
                var = ...
                class Cls:
                    var = ...
                    def inner_f():
                        nonlocal var
                        var = ...
            """,
            """
            def outer_f():
                import f as var
                class Cls:
                    var = ...
                    def inner_f():
                        nonlocal var
                        var = ...
            """,
            """
            def outer_f():
                var = ...
                class Cls:
                    var = ...
                    def inner_f():
                        nonlocal var
                        import f as var
            """,
        )
        for code in codes:
            m, scopes = get_scope_metadata_provider(code)
            scope_of_module = scopes[m]
            self.assertIsInstance(scope_of_module, GlobalScope)
            self.assertTrue("var" not in scope_of_module)

            outer_f = ensure_type(m.body[0], cst.FunctionDef)
            outer_f_body_var = ensure_type(
                ensure_type(outer_f.body.body[0], cst.SimpleStatementLine).body[0],
                cst.CSTNode,
            )
            scope_of_outer_f = scopes[outer_f_body_var]
            self.assertIsInstance(scope_of_outer_f, FunctionScope)
            self.assertTrue("var" in scope_of_outer_f)
            self.assertEqual(len(scope_of_outer_f["var"]), 2)

            cls = ensure_type(outer_f.body.body[1], cst.ClassDef)
            scope_of_cls = scopes[cls.body.body[0]]
            self.assertIsInstance(scope_of_cls, ClassScope)
            self.assertTrue("var" in scope_of_cls)

            inner_f = ensure_type(cls.body.body[1], cst.FunctionDef)
            inner_f_body_var = ensure_type(
                ensure_type(inner_f.body.body[1], cst.SimpleStatementLine).body[0],
                cst.CSTNode,
            )
            scope_of_inner_f = scopes[inner_f_body_var]
            self.assertIsInstance(scope_of_inner_f, FunctionScope)
            self.assertTrue("var" in scope_of_inner_f)
            self.assertEqual(len(scope_of_inner_f["var"]), 2)
            self.assertEqual(
                {
                    cast(Assignment, assignment).node
                    for assignment in scope_of_outer_f["var"]
                },
                {
                    outer_f_body_var.targets[0].target
                    if isinstance(outer_f_body_var, cst.Assign)
                    else outer_f_body_var,
                    inner_f_body_var.targets[0].target
                    if isinstance(inner_f_body_var, cst.Assign)
                    else inner_f_body_var,
                },
            )

    def test_local_scope_shadowing_with_functions(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            def f():
                def f():
                    f = ...
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        self.assertTrue("f" in scope_of_module)

        outer_f = ensure_type(m.body[0], cst.FunctionDef)
        scope_of_outer_f = scopes[outer_f.body.body[0]]
        self.assertIsInstance(scope_of_outer_f, FunctionScope)
        self.assertTrue("f" in scope_of_outer_f)
        out_f_assignment = list(scope_of_module["f"])[0]
        self.assertEqual(cast(Assignment, out_f_assignment).node, outer_f)

        inner_f = ensure_type(outer_f.body.body[0], cst.FunctionDef)
        scope_of_inner_f = scopes[inner_f.body.body[0]]
        self.assertIsInstance(scope_of_inner_f, FunctionScope)
        self.assertTrue("f" in scope_of_inner_f)
        inner_f_assignment = list(scope_of_outer_f["f"])[0]
        self.assertEqual(cast(Assignment, inner_f_assignment).node, inner_f)

    def test_func_param_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            @decorator
            def f(x: T=1, *vararg, y: T=2, z, **kwarg) -> RET:
                pass
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        self.assertTrue("f" in scope_of_module)

        f = ensure_type(m.body[0], cst.FunctionDef)
        scope_of_f = scopes[f.body.body[0]]
        self.assertIsInstance(scope_of_f, FunctionScope)

        decorator = f.decorators[0]
        x = f.params.params[0]
        xT = ensure_type(x.annotation, cst.Annotation)
        one = ensure_type(x.default, cst.BaseExpression)
        vararg = ensure_type(f.params.star_arg, cst.Param)
        y = f.params.kwonly_params[0]
        yT = ensure_type(y.annotation, cst.Annotation)
        two = ensure_type(y.default, cst.BaseExpression)
        z = f.params.kwonly_params[1]
        kwarg = ensure_type(f.params.star_kwarg, cst.Param)
        ret = ensure_type(f.returns, cst.Annotation).annotation

        self.assertEqual(scopes[decorator], scope_of_module)
        self.assertEqual(scopes[x], scope_of_f)
        self.assertEqual(scopes[xT], scope_of_module)
        self.assertEqual(scopes[one], scope_of_module)
        self.assertEqual(scopes[vararg], scope_of_f)
        self.assertEqual(scopes[y], scope_of_f)
        self.assertEqual(scopes[yT], scope_of_module)
        self.assertEqual(scopes[z], scope_of_f)
        self.assertEqual(scopes[two], scope_of_module)
        self.assertEqual(scopes[kwarg], scope_of_f)
        self.assertEqual(scopes[ret], scope_of_module)

        self.assertTrue("x" not in scope_of_module)
        self.assertTrue("x" in scope_of_f)
        self.assertTrue("vararg" not in scope_of_module)
        self.assertTrue("vararg" in scope_of_f)
        self.assertTrue("y" not in scope_of_module)
        self.assertTrue("y" in scope_of_f)
        self.assertTrue("z" not in scope_of_module)
        self.assertTrue("z" in scope_of_f)
        self.assertTrue("kwarg" not in scope_of_module)
        self.assertTrue("kwarg" in scope_of_f)

        self.assertEqual(cast(Assignment, list(scope_of_f["x"])[0]).node, x)
        self.assertEqual(cast(Assignment, list(scope_of_f["vararg"])[0]).node, vararg)
        self.assertEqual(cast(Assignment, list(scope_of_f["y"])[0]).node, y)
        self.assertEqual(cast(Assignment, list(scope_of_f["z"])[0]).node, z)
        self.assertEqual(cast(Assignment, list(scope_of_f["kwarg"])[0]).node, kwarg)

    def test_lambda_param_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            lambda x=1, *vararg, y=2, z, **kwarg:x
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)

        f = ensure_type(
            ensure_type(
                ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.Lambda,
        )
        scope_of_f = scopes[f.body]
        self.assertIsInstance(scope_of_f, FunctionScope)

        x = f.params.params[0]
        one = ensure_type(x.default, cst.BaseExpression)
        vararg = ensure_type(f.params.star_arg, cst.Param)
        y = f.params.kwonly_params[0]
        two = ensure_type(y.default, cst.BaseExpression)
        z = f.params.kwonly_params[1]
        kwarg = ensure_type(f.params.star_kwarg, cst.Param)

        self.assertEqual(scopes[x], scope_of_f)
        self.assertEqual(scopes[one], scope_of_module)
        self.assertEqual(scopes[vararg], scope_of_f)
        self.assertEqual(scopes[y], scope_of_f)
        self.assertEqual(scopes[z], scope_of_f)
        self.assertEqual(scopes[two], scope_of_module)
        self.assertEqual(scopes[kwarg], scope_of_f)

        self.assertTrue("x" not in scope_of_module)
        self.assertTrue("x" in scope_of_f)
        self.assertTrue("vararg" not in scope_of_module)
        self.assertTrue("vararg" in scope_of_f)
        self.assertTrue("y" not in scope_of_module)
        self.assertTrue("y" in scope_of_f)
        self.assertTrue("z" not in scope_of_module)
        self.assertTrue("z" in scope_of_f)
        self.assertTrue("kwarg" not in scope_of_module)
        self.assertTrue("kwarg" in scope_of_f)

        self.assertEqual(cast(Assignment, list(scope_of_f["x"])[0]).node, x)
        self.assertEqual(cast(Assignment, list(scope_of_f["vararg"])[0]).node, vararg)
        self.assertEqual(cast(Assignment, list(scope_of_f["y"])[0]).node, y)
        self.assertEqual(cast(Assignment, list(scope_of_f["z"])[0]).node, z)
        self.assertEqual(cast(Assignment, list(scope_of_f["kwarg"])[0]).node, kwarg)

    def test_except_handler(self) -> None:
        """
        The ``except as`` is a special case. The asname is only available in the excep body
        block and it'll be removed when existing the block.
        See https://docs.python.org/3.4/reference/compound_stmts.html#except
        We don't create a new block for except body because we don't handle del in our Scope Analysis.
        """
        m, scopes = get_scope_metadata_provider(
            """
            try:
                ...
            except Exception as ex:
                ...
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        self.assertTrue("ex" in scope_of_module)
        self.assertEqual(
            cast(Assignment, list(scope_of_module["ex"])[0]).node,
            ensure_type(
                ensure_type(m.body[0], cst.Try).handlers[0].name, cst.AsName
            ).name,
        )

    def test_with_asname(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            with open(file_name) as f:
                ...
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        self.assertTrue("f" in scope_of_module)
        self.assertEqual(
            cast(Assignment, list(scope_of_module["f"])[0]).node,
            ensure_type(
                ensure_type(m.body[0], cst.With).items[0].asname, cst.AsName
            ).name,
        )

    def test_get_qualified_names_for(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            from a.b import c
            class Cls:
                def f(self) -> "c":
                    c()
                    d = {}
                    d['key'] = 0
            def g():
                pass
            g()
            """
        )
        cls = ensure_type(m.body[1], cst.ClassDef)
        f = ensure_type(cls.body.body[0], cst.FunctionDef)
        scope_of_module = scopes[m]
        self.assertEqual(
            scope_of_module.get_qualified_names_for(
                ensure_type(f.returns, cst.Annotation).annotation
            ),
            set(),
            "Get qualified name given a SimpleString type annotation is not supported",
        )

        c_call = ensure_type(
            ensure_type(f.body.body[0], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        scope_of_f = scopes[c_call]
        self.assertIsInstance(scope_of_f, FunctionScope)
        self.assertEqual(
            scope_of_f.get_qualified_names_for(c_call),
            {QualifiedName("a.b.c", QualifiedNameSource.IMPORT)},
        )
        self.assertEqual(
            scope_of_f.get_qualified_names_for(c_call),
            {QualifiedName("a.b.c", QualifiedNameSource.IMPORT)},
        )

        g_call = ensure_type(
            ensure_type(m.body[3], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        self.assertIsInstance(scope_of_module, GlobalScope)
        self.assertEqual(
            scope_of_module.get_qualified_names_for(g_call),
            {QualifiedName("g", QualifiedNameSource.LOCAL)},
        )
        d_name = (
            ensure_type(
                ensure_type(f.body.body[1], cst.SimpleStatementLine).body[0], cst.Assign
            )
            .targets[0]
            .target
        )
        self.assertEqual(
            scope_of_f.get_qualified_names_for(d_name),
            {QualifiedName("Cls.f.<locals>.d", QualifiedNameSource.LOCAL)},
        )
        d_subscript = (
            ensure_type(
                ensure_type(f.body.body[2], cst.SimpleStatementLine).body[0], cst.Assign
            )
            .targets[0]
            .target
        )
        self.assertEqual(
            scope_of_f.get_qualified_names_for(d_subscript),
            {QualifiedName("Cls.f.<locals>.d", QualifiedNameSource.LOCAL)},
        )

        for builtin in ["map", "int", "dict"]:
            self.assertEqual(
                scope_of_f.get_qualified_names_for(cst.Name(value=builtin)),
                {QualifiedName(f"builtins.{builtin}", QualifiedNameSource.BUILTIN)},
                f"Test builtin: {builtin}.",
            )

        self.assertEqual(
            scope_of_module.get_qualified_names_for(cst.Name(value="d")),
            set(),
            "Test variable d in global scope.",
        )

    def test_get_qualified_names_for_nested_cases(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            class A:
                def f1(self):
                    def f2():
                        pass
                    f2()

                def f3(self):
                    class B():
                        ...
                    B()
            def f4():
                def f5():
                    class C:
                        pass
                    C()
                f5()
            """
        )
        cls_a = ensure_type(m.body[0], cst.ClassDef)
        func_f1 = ensure_type(cls_a.body.body[0], cst.FunctionDef)
        scope_of_cls_a = scopes[func_f1]
        self.assertIsInstance(scope_of_cls_a, ClassScope)
        self.assertEqual(
            scope_of_cls_a.get_qualified_names_for(func_f1),
            {QualifiedName("A.f1", QualifiedNameSource.LOCAL)},
        )
        func_f2_call = ensure_type(
            ensure_type(func_f1.body.body[1], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        scope_of_f1 = scopes[func_f2_call]
        self.assertIsInstance(scope_of_f1, FunctionScope)
        self.assertEqual(
            scope_of_f1.get_qualified_names_for(func_f2_call),
            {QualifiedName("A.f1.<locals>.f2", QualifiedNameSource.LOCAL)},
        )
        func_f3 = ensure_type(cls_a.body.body[1], cst.FunctionDef)
        call_b = ensure_type(
            ensure_type(func_f3.body.body[1], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        scope_of_f3 = scopes[call_b]
        self.assertIsInstance(scope_of_f3, FunctionScope)
        self.assertEqual(
            scope_of_f3.get_qualified_names_for(call_b),
            {QualifiedName("A.f3.<locals>.B", QualifiedNameSource.LOCAL)},
        )
        func_f4 = ensure_type(m.body[1], cst.FunctionDef)
        func_f5 = ensure_type(func_f4.body.body[0], cst.FunctionDef)
        scope_of_f4 = scopes[func_f5]
        self.assertIsInstance(scope_of_f4, FunctionScope)
        self.assertEqual(
            scope_of_f4.get_qualified_names_for(func_f5),
            {QualifiedName("f4.<locals>.f5", QualifiedNameSource.LOCAL)},
        )
        cls_c = func_f5.body.body[0]
        scope_of_f5 = scopes[cls_c]
        self.assertIsInstance(scope_of_f5, FunctionScope)
        self.assertEqual(
            scope_of_f5.get_qualified_names_for(cls_c),
            {QualifiedName("f4.<locals>.f5.<locals>.C", QualifiedNameSource.LOCAL)},
        )

    def test_get_qualified_names_for_the_same_prefix(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                from a import b, bc
                bc()
            """
        )
        call = ensure_type(
            ensure_type(
                ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.Call,
        )
        module_scope = scopes[m]
        self.assertEqual(
            module_scope.get_qualified_names_for(call.func),
            {QualifiedName("a.bc", QualifiedNameSource.IMPORT)},
        )

    def test_get_qualified_names_for_dotted_imports(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                import a.b.c
                a(a.b.d)
            """
        )
        call = ensure_type(
            ensure_type(
                ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.Call,
        )
        module_scope = scopes[m]
        self.assertEqual(
            module_scope.get_qualified_names_for(call.func),
            {QualifiedName("a", QualifiedNameSource.IMPORT)},
        )
        self.assertEqual(
            module_scope.get_qualified_names_for(call.args[0].value),
            {QualifiedName("a.b.d", QualifiedNameSource.IMPORT)},
        )

        import_stmt = ensure_type(
            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.Import
        )
        a_b_c = ensure_type(import_stmt.names[0].name, cst.Attribute)
        a_b = ensure_type(a_b_c.value, cst.Attribute)
        a = a_b.value
        self.assertEqual(
            module_scope.get_qualified_names_for(a_b_c),
            {QualifiedName("a.b.c", QualifiedNameSource.IMPORT)},
        )
        self.assertEqual(
            module_scope.get_qualified_names_for(a_b),
            {QualifiedName("a.b", QualifiedNameSource.IMPORT)},
        )
        self.assertEqual(
            module_scope.get_qualified_names_for(a),
            {QualifiedName("a", QualifiedNameSource.IMPORT)},
        )

    def test_multiple_assignments(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                if 1:
                    from a import b as c
                elif 2:
                    from d import e as c
                c()
            """
        )
        call = ensure_type(
            ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        scope = scopes[call]
        self.assertIsInstance(scope, GlobalScope)
        self.assertEqual(
            scope.get_qualified_names_for(call),
            {
                QualifiedName(name="a.b", source=QualifiedNameSource.IMPORT),
                QualifiedName(name="d.e", source=QualifiedNameSource.IMPORT),
            },
        )
        self.assertEqual(
            scope.get_qualified_names_for("c"),
            {
                QualifiedName(name="a.b", source=QualifiedNameSource.IMPORT),
                QualifiedName(name="d.e", source=QualifiedNameSource.IMPORT),
            },
        )

    def test_assignments_and_accesses(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                a = 1
                def f():
                    a = 2
                    a, b
                    def g():
                        b = a
                a
            """
        )
        a_outer_assign = (
            ensure_type(
                ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.Assign
            )
            .targets[0]
            .target
        )
        a_outer_access = ensure_type(
            ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        scope_of_module = scopes[a_outer_assign]
        a_outer_assignments = scope_of_module.assignments[a_outer_access]
        self.assertEqual(len(a_outer_assignments), 1)
        a_outer_assignment = list(a_outer_assignments)[0]
        self.assertEqual(cast(Assignment, a_outer_assignment).node, a_outer_assign)
        self.assertEqual(
            {i.node for i in a_outer_assignment.references}, {a_outer_access}
        )

        a_outer_assesses = scope_of_module.accesses[a_outer_assign]
        self.assertEqual(len(a_outer_assesses), 1)
        self.assertEqual(list(a_outer_assesses)[0].node, a_outer_access)

        self.assertEqual(
            {cast(Assignment, i).node for i in list(a_outer_assesses)[0].referents},
            {a_outer_assign},
        )

        self.assertTrue(a_outer_assign in scope_of_module.accesses)
        self.assertTrue(a_outer_assign in scope_of_module.assignments)
        self.assertTrue(a_outer_access in scope_of_module.accesses)
        self.assertTrue(a_outer_access in scope_of_module.assignments)

        f = ensure_type(m.body[1], cst.FunctionDef)
        a_inner_assign = (
            ensure_type(
                ensure_type(
                    ensure_type(f.body, cst.IndentedBlock).body[0],
                    cst.SimpleStatementLine,
                ).body[0],
                cst.Assign,
            )
            .targets[0]
            .target
        )
        scope_of_f = scopes[a_inner_assign]
        a_inner_assignments = scope_of_f.assignments["a"]
        self.assertEqual(len(a_inner_assignments), 1)
        self.assertEqual(
            cast(Assignment, list(a_inner_assignments)[0]).node, a_inner_assign
        )
        tup = ensure_type(
            ensure_type(
                ensure_type(
                    ensure_type(f.body, cst.IndentedBlock).body[1],
                    cst.SimpleStatementLine,
                ).body[0],
                cst.Expr,
            ).value,
            cst.Tuple,
        )
        a_inner_access = tup.elements[0].value
        b_inner_access = tup.elements[1].value
        all_inner_accesses = [i for i in scope_of_f.accesses]
        self.assertEqual(len(all_inner_accesses), 2)
        self.assertEqual(
            {i.node for i in all_inner_accesses}, {a_inner_access, b_inner_access}
        )

        g = ensure_type(ensure_type(f.body, cst.IndentedBlock).body[2], cst.FunctionDef)
        inner_most_assign = ensure_type(
            ensure_type(g.body.body[0], cst.SimpleStatementLine).body[0], cst.Assign
        )
        b_inner_most_assign = inner_most_assign.targets[0].target
        a_inner_most_access = inner_most_assign.value
        scope_of_g = scopes[b_inner_most_assign]
        self.assertEqual({i.node for i in scope_of_g.accesses}, {a_inner_most_access})
        self.assertEqual(
            {cast(Assignment, i).node for i in scope_of_g.assignments},
            {b_inner_most_assign},
        )

        self.assertEqual(len(set(scopes.values())), 3)

    def test_annotation_access(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                from typing import Literal, NewType, Optional, TypeVar, Callable, cast
                from a import A, B, C, D, D2, E, E2, F, G, G2, H, I, J, K, K2, L, M
                def x(a: A):
                    pass
                def y(b: "B"):
                    pass
                def z(c: Literal["C"]):
                    pass
                DType = TypeVar("D2", bound=D)
                EType = TypeVar("E2", bound="E")
                FType = TypeVar("F")
                GType = NewType("G2", "Optional[G]")
                HType = Optional["H"]
                IType = Callable[..., I]

                class Test(Generic[J]):
                    pass
                castedK = cast("K", "K2")
                castedL = cast("L", M)
            """
        )
        imp = ensure_type(
            ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.ImportFrom
        )
        scope = scopes[imp]

        assignment = list(scope["A"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertTrue(references[0].is_annotation)

        assignment = list(scope["B"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertTrue(references[0].is_annotation)
        reference_node = references[0].node
        self.assertIsInstance(reference_node, cst.SimpleString)
        if isinstance(reference_node, cst.SimpleString):
            self.assertEqual(reference_node.evaluated_value, "B")

        assignment = list(scope["C"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 0)

        assignment = list(scope["D"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertFalse(references[0].is_annotation)
        self.assertTrue(references[0].is_type_hint)

        assignment = list(scope["D2"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 0)

        assignment = list(scope["E"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertFalse(references[0].is_annotation)
        self.assertTrue(references[0].is_type_hint)
        reference_node = references[0].node
        self.assertIsInstance(reference_node, cst.SimpleString)
        if isinstance(reference_node, cst.SimpleString):
            self.assertEqual(reference_node.evaluated_value, "E")

        assignment = list(scope["E2"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 0)

        assignment = list(scope["F"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 0)

        assignment = list(scope["G"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertFalse(references[0].is_annotation)
        self.assertTrue(references[0].is_type_hint)
        reference_node = references[0].node
        self.assertIsInstance(reference_node, cst.SimpleString)
        if isinstance(reference_node, cst.SimpleString):
            self.assertEqual(reference_node.evaluated_value, "Optional[G]")

        assignment = list(scope["G2"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 0)

        assignment = list(scope["H"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertFalse(references[0].is_annotation)
        self.assertTrue(references[0].is_type_hint)
        reference_node = references[0].node
        self.assertIsInstance(reference_node, cst.SimpleString)
        if isinstance(reference_node, cst.SimpleString):
            self.assertEqual(reference_node.evaluated_value, "H")

        assignment = list(scope["I"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertFalse(references[0].is_annotation)

        assignment = list(scope["J"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertFalse(references[0].is_annotation)

        assignment = list(scope["K"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertFalse(references[0].is_annotation)
        reference_node = references[0].node
        self.assertIsInstance(reference_node, cst.SimpleString)
        if isinstance(reference_node, cst.SimpleString):
            self.assertEqual(reference_node.evaluated_value, "K")

        assignment = list(scope["K2"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 0)

        assignment = list(scope["L"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        reference_node = references[0].node
        self.assertIsInstance(reference_node, cst.SimpleString)
        if isinstance(reference_node, cst.SimpleString):
            self.assertEqual(reference_node.evaluated_value, "L")

        assignment = list(scope["M"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)

    def test_insane_annotation_access(self) -> None:
        m, scopes = get_scope_metadata_provider(
            r"""
                from typing import TypeVar, Optional
                from a import G
                TypeVar("G2", bound="Optional[\"G\"]")
            """
        )
        imp = ensure_type(
            ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.ImportFrom
        )
        call = ensure_type(
            ensure_type(
                ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.Call,
        )
        bound = call.args[1].value
        scope = scopes[imp]
        assignment = next(iter(scope["G"]))
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        self.assertEqual(list(assignment.references)[0].node, bound)

    def test_dotted_annotation_access(self) -> None:
        m, scopes = get_scope_metadata_provider(
            r"""
                from typing import TypeVar
                import a.G
                TypeVar("G2", bound="a.G")
            """
        )
        imp = ensure_type(
            ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Import
        )
        call = ensure_type(
            ensure_type(
                ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Expr
            ).value,
            cst.Call,
        )
        bound = call.args[1].value
        scope = scopes[imp]
        assignment = next(iter(scope["a.G"]))
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        self.assertEqual(list(assignment.references)[0].node, bound)

    def test_node_of_scopes(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                def f1():
                    target()

                class C:
                    attr = target()
            """
        )
        f1 = ensure_type(m.body[0], cst.FunctionDef)
        target_call = ensure_type(
            ensure_type(f1.body.body[0], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        f1_scope = scopes[target_call]
        self.assertIsInstance(f1_scope, FunctionScope)
        self.assertEqual(cast(FunctionScope, f1_scope).node, f1)
        c = ensure_type(m.body[1], cst.ClassDef)
        target_call_2 = ensure_type(
            ensure_type(c.body.body[0], cst.SimpleStatementLine).body[0], cst.Assign
        ).value
        c_scope = scopes[target_call_2]
        self.assertIsInstance(c_scope, ClassScope)
        self.assertEqual(cast(ClassScope, c_scope).node, c)

    def test_with_statement(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                import unittest.mock

                with unittest.mock.patch("something") as obj:
                    obj.f1()

                unittest.mock
            """
        )
        import_ = ensure_type(m.body[0], cst.SimpleStatementLine).body[0]
        assignments = scopes[import_]["unittest"]
        self.assertEqual(len(assignments), 1)
        self.assertEqual(cast(Assignment, list(assignments)[0]).node, import_)
        with_ = ensure_type(m.body[1], cst.With)
        fn_call = with_.items[0].item
        self.assertEqual(
            scopes[fn_call].get_qualified_names_for(fn_call),
            {
                QualifiedName(
                    name="unittest.mock.patch", source=QualifiedNameSource.IMPORT
                )
            },
        )
        mock = ensure_type(
            ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        self.assertEqual(
            scopes[fn_call].get_qualified_names_for(mock),
            {QualifiedName(name="unittest.mock", source=QualifiedNameSource.IMPORT)},
        )

    def test_del_context_names(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                import a
                dic = {}
                del dic
                del dic["key"]
                del a.b
            """
        )
        dic = ensure_type(
            ensure_type(
                ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Assign
            ).targets[0],
            cst.AssignTarget,
        ).target
        del_dic = ensure_type(
            ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Del
        )
        scope = scopes[del_dic]
        assignments = list(scope["dic"])
        self.assertEqual(len(assignments), 1)
        dic_assign = assignments[0]
        self.assertIsInstance(dic_assign, Assignment)
        self.assertEqual(cast(Assignment, dic_assign).node, dic)
        self.assertEqual(len(dic_assign.references), 2)
        del_dic_subscript = ensure_type(
            ensure_type(
                ensure_type(m.body[3], cst.SimpleStatementLine).body[0], cst.Del
            ).target,
            cst.Subscript,
        )
        self.assertSetEqual(
            {i.node for i in dic_assign.references},
            {del_dic.target, del_dic_subscript.value},
        )
        assignments = list(scope["a"])
        self.assertEqual(len(assignments), 1)
        a_assign = assignments[0]
        self.assertIsInstance(a_assign, Assignment)
        import_a = ensure_type(m.body[0], cst.SimpleStatementLine).body[0]
        self.assertEqual(cast(Assignment, a_assign).node, import_a)
        self.assertEqual(len(a_assign.references), 1)
        del_a_b = ensure_type(
            ensure_type(m.body[4], cst.SimpleStatementLine).body[0], cst.Del
        )
        self.assertEqual(
            {i.node for i in a_assign.references},
            {ensure_type(del_a_b.target, cst.Attribute).value},
        )
        self.assertEqual(scope["b"], set())

    def test_keyword_arg_in_call(self) -> None:
        m, scopes = get_scope_metadata_provider("call(arg=val)")
        call = ensure_type(
            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.Expr
        ).value
        scope = scopes[call]
        self.assertIsInstance(scope, GlobalScope)
        self.assertEqual(len(scope["arg"]), 0)  # no assignment should exist

    def test_global_contains_is_read_only(self) -> None:
        gscope = GlobalScope()
        before_assignments = list(gscope.assignments)
        before_accesses = list(gscope.accesses)
        self.assertFalse("doesnt_exist" in gscope)
        self.assertEqual(list(gscope.accesses), before_accesses)
        self.assertEqual(list(gscope.assignments), before_assignments)

    def test_contains_is_read_only(self) -> None:
        for s in [LocalScope, FunctionScope, ClassScope, ComprehensionScope]:
            with self.subTest(scope=s):
                gscope = GlobalScope()
                scope = s(parent=gscope, node=cst.Name("lol"))
                before_assignments = list(scope.assignments)
                before_accesses = list(scope.accesses)
                before_overwrites = list(scope._scope_overwrites.items())
                before_parent_assignments = list(scope.parent.assignments)
                before_parent_accesses = list(scope.parent.accesses)

                self.assertFalse("doesnt_exist" in scope)
                self.assertEqual(list(scope.accesses), before_accesses)
                self.assertEqual(list(scope.assignments), before_assignments)
                self.assertEqual(
                    list(scope._scope_overwrites.items()), before_overwrites
                )
                self.assertEqual(
                    list(scope.parent.assignments), before_parent_assignments
                )
                self.assertEqual(list(scope.parent.accesses), before_parent_accesses)

    def test_attribute_of_function_call(self) -> None:
        get_scope_metadata_provider("foo().bar")

    def test_attribute_of_subscript_called(self) -> None:
        m, scopes = get_scope_metadata_provider("foo[0].bar.baz()")
        scope = scopes[m]
        self.assertIn("foo", scope.accesses)

    def test_self(self) -> None:
        with open(__file__) as f:
            get_scope_metadata_provider(f.read())

    def test_get_qualified_names_for_is_read_only(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                import a
                import b
            """
        )
        a = m.body[0]
        scope = scopes[a]
        assignments_before = list(scope.assignments)
        accesses_before = list(scope.accesses)
        scope.get_qualified_names_for("doesnt_exist")
        self.assertEqual(list(scope.assignments), assignments_before)
        self.assertEqual(list(scope.accesses), accesses_before)

    def test_gen_dotted_names(self) -> None:
        names = {name for name, node in _gen_dotted_names(cst.Name(value="a"))}
        self.assertEqual(names, {"a"})

        names = {
            name
            for name, node in _gen_dotted_names(
                cst.Attribute(value=cst.Name(value="a"), attr=cst.Name(value="b"))
            )
        }
        self.assertEqual(names, {"a.b", "a"})

        names = {
            name
            for name, node in _gen_dotted_names(
                cst.Attribute(
                    value=cst.Call(
                        func=cst.Attribute(
                            value=cst.Attribute(
                                value=cst.Name(value="a"), attr=cst.Name(value="b")
                            ),
                            attr=cst.Name(value="c"),
                        ),
                        args=[],
                    ),
                    attr=cst.Name(value="d"),
                )
            )
        }
        self.assertEqual(names, {"a.b.c", "a.b", "a"})

    def test_ordering(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            from a import b
            class X:
                x = b
                b = b
                y = b
            """
        )
        global_scope = scopes[m]
        import_stmt = ensure_type(
            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.ImportFrom
        )
        first_assignment = list(global_scope.assignments)[0]
        assert isinstance(first_assignment, cst.metadata.Assignment)
        self.assertEqual(first_assignment.node, import_stmt)
        global_refs = first_assignment.references
        self.assertEqual(len(global_refs), 2)
        global_refs_nodes = {x.node for x in global_refs}
        class_def = ensure_type(m.body[1], cst.ClassDef)
        x = ensure_type(
            ensure_type(class_def.body.body[0], cst.SimpleStatementLine).body[0],
            cst.Assign,
        )
        self.assertIn(x.value, global_refs_nodes)
        class_b = ensure_type(
            ensure_type(class_def.body.body[1], cst.SimpleStatementLine).body[0],
            cst.Assign,
        )
        self.assertIn(class_b.value, global_refs_nodes)

        class_accesses = list(scopes[x].accesses)
        self.assertEqual(len(class_accesses), 3)
        self.assertIn(
            class_b.targets[0].target,
            [
                ref.node
                for acc in class_accesses
                for ref in acc.referents
                if isinstance(ref, Assignment)
            ],
        )
        y = ensure_type(
            ensure_type(class_def.body.body[2], cst.SimpleStatementLine).body[0],
            cst.Assign,
        )
        self.assertIn(y.value, [access.node for access in class_accesses])

    def test_ordering_between_scopes(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            def f(a):
                print(a)
                print(b)
            a = 1
            b = 1
            """
        )
        f = cst.ensure_type(m.body[0], cst.FunctionDef)
        a_param = f.params.params[0].name
        a_param_assignment = list(scopes[a_param]["a"])[0]
        a_param_refs = list(a_param_assignment.references)
        first_print = cst.ensure_type(
            cst.ensure_type(
                cst.ensure_type(f.body.body[0], cst.SimpleStatementLine).body[0],
                cst.Expr,
            ).value,
            cst.Call,
        )
        second_print = cst.ensure_type(
            cst.ensure_type(
                cst.ensure_type(f.body.body[1], cst.SimpleStatementLine).body[0],
                cst.Expr,
            ).value,
            cst.Call,
        )
        self.assertEqual(
            first_print.args[0].value,
            a_param_refs[0].node,
        )
        a_global = (
            cst.ensure_type(
                cst.ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Assign
            )
            .targets[0]
            .target
        )
        a_global_assignment = list(scopes[a_global]["a"])[0]
        a_global_refs = list(a_global_assignment.references)
        self.assertEqual(a_global_refs, [])
        b_global = (
            cst.ensure_type(
                cst.ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.Assign
            )
            .targets[0]
            .target
        )
        b_global_assignment = list(scopes[b_global]["b"])[0]
        b_global_refs = list(b_global_assignment.references)
        self.assertEqual(len(b_global_refs), 1)
        self.assertEqual(b_global_refs[0].node, second_print.args[0].value)

    def test_ordering_comprehension(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            def f(a):
                [a for a in [] for b in a]
                [b for a in [] for b in a]
                [a for a in [] for a in []]
            a = 1
            """
        )
        f = cst.ensure_type(m.body[0], cst.FunctionDef)
        a_param = f.params.params[0].name
        a_param_assignment = list(scopes[a_param]["a"])[0]
        a_param_refs = list(a_param_assignment.references)
        self.assertEqual(a_param_refs, [])
        first_comp = cst.ensure_type(
            cst.ensure_type(
                cst.ensure_type(f.body.body[0], cst.SimpleStatementLine).body[0],
                cst.Expr,
            ).value,
            cst.ListComp,
        )
        a_comp_assignment = list(scopes[first_comp.elt]["a"])[0]
        self.assertEqual(len(a_comp_assignment.references), 2)
        self.assertIn(
            first_comp.elt, [ref.node for ref in a_comp_assignment.references]
        )

        second_comp = cst.ensure_type(
            cst.ensure_type(
                cst.ensure_type(f.body.body[1], cst.SimpleStatementLine).body[0],
                cst.Expr,
            ).value,
            cst.ListComp,
        )
        b_comp_assignment = list(scopes[second_comp.elt]["b"])[0]
        self.assertEqual(len(b_comp_assignment.references), 1)
        a_second_comp_assignment = list(scopes[second_comp.elt]["a"])[0]
        self.assertEqual(len(a_second_comp_assignment.references), 1)

        third_comp = cst.ensure_type(
            cst.ensure_type(
                cst.ensure_type(f.body.body[2], cst.SimpleStatementLine).body[0],
                cst.Expr,
            ).value,
            cst.ListComp,
        )
        a_third_comp_assignments = list(scopes[third_comp.elt]["a"])
        self.assertEqual(len(a_third_comp_assignments), 2)
        a_third_comp_access = list(scopes[third_comp.elt].accesses)[0]
        self.assertEqual(a_third_comp_access.node, third_comp.elt)
        # We record both assignments because it's impossible to know which one
        # the access refers to without running the program
        self.assertEqual(len(a_third_comp_access.referents), 2)
        inner_for_in = third_comp.for_in.inner_for_in
        self.assertIsNotNone(inner_for_in)
        if inner_for_in:
            self.assertIn(
                inner_for_in.target,
                {
                    ref.node
                    for ref in a_third_comp_access.referents
                    if isinstance(ref, Assignment)
                },
            )

        a_global = (
            cst.ensure_type(
                cst.ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.Assign
            )
            .targets[0]
            .target
        )
        a_global_assignment = list(scopes[a_global]["a"])[0]
        a_global_refs = list(a_global_assignment.references)
        self.assertEqual(a_global_refs, [])

    def test_ordering_comprehension_confusing(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            def f(a):
                [a for a in a]
            a = 1
            """
        )
        f = cst.ensure_type(m.body[0], cst.FunctionDef)
        a_param = f.params.params[0].name
        a_param_assignment = list(scopes[a_param]["a"])[0]
        a_param_refs = list(a_param_assignment.references)
        self.assertEqual(len(a_param_refs), 1)
        comp = cst.ensure_type(
            cst.ensure_type(
                cst.ensure_type(f.body.body[0], cst.SimpleStatementLine).body[0],
                cst.Expr,
            ).value,
            cst.ListComp,
        )
        a_comp_assignment = list(scopes[comp.elt]["a"])[0]
        self.assertEqual(list(a_param_refs)[0].node, comp.for_in.iter)
        self.assertEqual(len(a_comp_assignment.references), 1)
        self.assertEqual(list(a_comp_assignment.references)[0].node, comp.elt)

    def test_for_scope_ordering(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            def f():
                for x in []:
                    x
            class X:
                def f():
                    for x in []:
                        x
            """
        )
        for scope in scopes.values():
            for acc in scope.accesses:
                self.assertEqual(
                    len(acc.referents),
                    1,
                    msg=(
                        "Access for node has incorrect number of referents: "
                        + f"{acc.node}"
                    ),
                )

    def test_no_out_of_order_references_in_global_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            x = y
            y = 1
            """
        )
        for scope in scopes.values():
            for acc in scope.accesses:
                self.assertEqual(
                    len(acc.referents),
                    0,
                    msg=(
                        "Access for node has incorrect number of referents: "
                        + f"{acc.node}"
                    ),
                )

    def test_walrus_accesses(self) -> None:
        if sys.version_info < (3, 8):
            self.skipTest("This python version doesn't support :=")
        m, scopes = get_scope_metadata_provider(
            """
            if x := y:
                y = 1
                x
            """
        )
        for scope in scopes.values():
            for acc in scope.accesses:
                self.assertEqual(
                    len(acc.referents),
                    1 if getattr(acc.node, "value", None) == "x" else 0,
                    msg=(
                        "Access for node has incorrect number of referents: "
                        + f"{acc.node}"
                    ),
                )

    @data_provider(
        {
            "TypeVar": {
                "code": """
                    from typing import TypeVar
                    TypeVar("Name", "int")
                """,
                "calls": [mock.call("int")],
            },
            "Dict": {
                "code": """
                    from typing import Dict
                    Dict["str", "int"]
                """,
                "calls": [mock.call("str"), mock.call("int")],
            },
            "cast_no_annotation": {
                "code": """
                    from typing import Dict, cast
                    cast(Dict[str, str], {})["3rr0r"]
                """,
                "calls": [],
            },
            "cast_second_arg": {
                "code": """
                    from typing import cast
                    cast(str, "foo")
                """,
                "calls": [],
            },
            "cast_first_arg": {
                "code": """
                    from typing import cast
                    cast("int", "foo")
                """,
                "calls": [
                    mock.call("int"),
                ],
            },
            "typevar_func": {
                "code": """
                    from typing import TypeVar
                    TypeVar("Name", func("int"))
                """,
                "calls": [],
            },
            "literal": {
                "code": """
                    from typing import Literal
                    Literal[\"G\"]
                """,
                "calls": [],
            },
            "nested_str": {
                "code": r"""
                    from typing import TypeVar, Optional
                    from a import G
                    TypeVar("G2", bound="Optional[\"G\"]")
                """,
                "calls": [mock.call('Optional["G"]'), mock.call("G")],
            },
            "class_self_ref": {
                "code": """
                    from typing import TypeVar
                    class HelperClass:
                        value: TypeVar("THelperClass", bound="HelperClass")
                """,
                "calls": [mock.call("HelperClass")],
            },
        }
    )
    def test_parse_string_annotations(
        self, *, code: str, calls: Sequence[mock._Call]
    ) -> None:
        parse = cst.parse_module
        with mock.patch("libcst.parse_module") as parse_mock:
            parse_mock.side_effect = parse
            get_scope_metadata_provider(dedent(code))
            calls = [mock.call(dedent(code))] + list(calls)
            self.assertEqual(parse_mock.call_count, len(calls))
            parse_mock.assert_has_calls(calls)

    def test_builtin_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            a = pow(1, 2)
            def foo():
                b = pow(2, 3)
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        self.assertEqual(len(scope_of_module["pow"]), 1)
        builtin_pow_assignment = list(scope_of_module["pow"])[0]
        self.assertIsInstance(builtin_pow_assignment, BuiltinAssignment)
        self.assertIsInstance(builtin_pow_assignment.scope, BuiltinScope)

        global_a_assignments = scope_of_module["a"]
        self.assertEqual(len(global_a_assignments), 1)
        a_assignment = list(global_a_assignments)[0]
        self.assertIsInstance(a_assignment, Assignment)

        func_body = ensure_type(m.body[1], cst.FunctionDef).body
        func_statement = func_body.body[0]
        scope_of_func_statement = scopes[func_statement]
        self.assertIsInstance(scope_of_func_statement, FunctionScope)
        func_b_assignments = scope_of_func_statement["b"]
        self.assertEqual(len(func_b_assignments), 1)
        b_assignment = list(func_b_assignments)[0]
        self.assertIsInstance(b_assignment, Assignment)

        builtin_pow_accesses = list(builtin_pow_assignment.references)
        self.assertEqual(len(builtin_pow_accesses), 2)

    def test_override_builtin_scope(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
            def pow(x, y):
                return x ** y

            a = pow(1, 2)
            def foo():
                b = pow(2, 3)
            """
        )
        scope_of_module = scopes[m]
        self.assertIsInstance(scope_of_module, GlobalScope)
        self.assertEqual(len(scope_of_module["pow"]), 1)
        global_pow_assignment = list(scope_of_module["pow"])[0]
        self.assertIsInstance(global_pow_assignment, Assignment)
        self.assertIsInstance(global_pow_assignment.scope, GlobalScope)

        global_a_assignments = scope_of_module["a"]
        self.assertEqual(len(global_a_assignments), 1)
        a_assignment = list(global_a_assignments)[0]
        self.assertIsInstance(a_assignment, Assignment)

        func_body = ensure_type(m.body[2], cst.FunctionDef).body
        func_statement = func_body.body[0]
        scope_of_func_statement = scopes[func_statement]
        self.assertIsInstance(scope_of_func_statement, FunctionScope)
        func_b_assignments = scope_of_func_statement["b"]
        self.assertEqual(len(func_b_assignments), 1)
        b_assignment = list(func_b_assignments)[0]
        self.assertIsInstance(b_assignment, Assignment)

        global_pow_accesses = list(global_pow_assignment.references)
        self.assertEqual(len(global_pow_accesses), 2)

    def test_annotation_access_in_typevar_bound(self) -> None:
        m, scopes = get_scope_metadata_provider(
            """
                from typing import TypeVar
                class Test:
                    var: TypeVar("T", bound="Test")
            """
        )
        imp = ensure_type(
            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.ImportFrom
        )
        scope = scopes[imp]
        assignment = list(scope["Test"])[0]
        self.assertIsInstance(assignment, Assignment)
        self.assertEqual(len(assignment.references), 1)
        references = list(assignment.references)
        self.assertTrue(references[0].is_annotation)

    def test_prefix_match(self) -> None:
        """Verify that a name doesn't overmatch on prefix"""
        m, scopes = get_scope_metadata_provider(
            """
                def something():
                    ...
            """
        )
        scope = scopes[m]
        self.assertEqual(
            scope.get_qualified_names_for(cst.Name("something")),
            {QualifiedName(name="something", source=QualifiedNameSource.LOCAL)},
        )
        self.assertEqual(
            scope.get_qualified_names_for(cst.Name("something_else")),
            set(),
        )

    def test_type_alias_scope(self) -> None:
        if not is_native():
            self.skipTest("type aliases are only supported in the native parser")
        m, scopes = get_scope_metadata_provider(
            """
                type A = C
                lol: A
            """
        )
        alias = ensure_type(
            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.TypeAlias
        )
        self.assertIsInstance(scopes[alias], GlobalScope)
        a_assignments = list(scopes[alias]["A"])
        self.assertEqual(len(a_assignments), 1)
        lol = ensure_type(
            ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.AnnAssign
        )
        self.assertEqual(len(a_references := list(a_assignments[0].references)), 1)
        self.assertEqual(a_references[0].node, lol.annotation.annotation)

        self.assertIsInstance(scopes[alias.value], AnnotationScope)

    def test_type_alias_param(self) -> None:
        if not is_native():
            self.skipTest("type parameters are only supported in the native parser")
        m, scopes = get_scope_metadata_provider(
            """
                B = int
                type A[T: B] = T
                lol: T
            """
        )
        alias = ensure_type(
            ensure_type(m.body[1], cst.SimpleStatementLine).body[0], cst.TypeAlias
        )
        assert alias.type_parameters
        param_scope = scopes[alias.type_parameters]
        self.assertEqual(len(t_assignments := list(param_scope["T"])), 1)
        self.assertEqual(len(t_refs := list(t_assignments[0].references)), 1)
        self.assertEqual(t_refs[0].node, alias.value)

        b = (
            ensure_type(
                ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.Assign
            )
            .targets[0]
            .target
        )
        b_assignment = list(scopes[b]["B"])[0]
        self.assertEqual(
            {ref.node for ref in b_assignment.references},
            {ensure_type(alias.type_parameters.params[0].param, cst.TypeVar).bound},
        )

    def test_type_alias_tuple_and_paramspec(self) -> None:
        if not is_native():
            self.skipTest("type parameters are only supported in the native parser")
        m, scopes = get_scope_metadata_provider(
            """
            type A[*T] = T
            lol: T
            type A[**T] = T
            lol: T
            """
        )
        alias_tuple = ensure_type(
            ensure_type(m.body[0], cst.SimpleStatementLine).body[0], cst.TypeAlias
        )
        assert alias_tuple.type_parameters
        param_scope = scopes[alias_tuple.type_parameters]
        self.assertEqual(len(t_assignments := list(param_scope["T"])), 1)
        self.assertEqual(len(t_refs := list(t_assignments[0].references)), 1)
        self.assertEqual(t_refs[0].node, alias_tuple.value)

        alias_paramspec = ensure_type(
            ensure_type(m.body[2], cst.SimpleStatementLine).body[0], cst.TypeAlias
        )
        assert alias_paramspec.type_parameters
        param_scope = scopes[alias_paramspec.type_parameters]
        self.assertEqual(len(t_assignments := list(param_scope["T"])), 1)
        self.assertEqual(len(t_refs := list(t_assignments[0].references)), 1)
        self.assertEqual(t_refs[0].node, alias_paramspec.value)

    def test_class_type_params(self) -> None:
        if not is_native():
            self.skipTest("type parameters are only supported in the native parser")
        m, scopes = get_scope_metadata_provider(
            """
            class W[T]:
                def f() -> T: pass
                def g[T]() -> T: pass
            """
        )
        cls = ensure_type(m.body[0], cst.ClassDef)
        cls_scope = scopes[cls.body.body[0]]
        self.assertEqual(len(t_assignments_in_cls := list(cls_scope["T"])), 1)
        assert cls.type_parameters
        self.assertEqual(
            ensure_type(t_assignments_in_cls[0], Assignment).node,
            cls.type_parameters.params[0].param,
        )
        self.assertEqual(
            len(t_refs_in_cls := list(t_assignments_in_cls[0].references)), 1
        )
        f = ensure_type(cls.body.body[0], cst.FunctionDef)
        assert f.returns
        self.assertEqual(t_refs_in_cls[0].node, f.returns.annotation)

        g = ensure_type(cls.body.body[1], cst.FunctionDef)
        assert g.type_parameters
        assert g.returns
        self.assertEqual(len(t_assignments_in_g := list(scopes[g.body]["T"])), 1)
        self.assertEqual(
            ensure_type(t_assignments_in_g[0], Assignment).node,
            g.type_parameters.params[0].param,
        )
        self.assertEqual(len(t_refs_in_g := list(t_assignments_in_g[0].references)), 1)
        self.assertEqual(t_refs_in_g[0].node, g.returns.annotation)

    def test_nested_class_type_params(self) -> None:
        if not is_native():
            self.skipTest("type parameters are only supported in the native parser")
        m, scopes = get_scope_metadata_provider(
            """
            class Outer:
                class Nested[T: Outer]: pass
            """
        )
        outer = ensure_type(m.body[0], cst.ClassDef)
        outer_refs = list(list(scopes[outer]["Outer"])[0].references)
        self.assertEqual(len(outer_refs), 1)
        inner = ensure_type(outer.body.body[0], cst.ClassDef)
        assert inner.type_parameters
        self.assertEqual(
            outer_refs[0].node,
            ensure_type(inner.type_parameters.params[0].param, cst.TypeVar).bound,
        )

    def test_annotation_refers_to_nested_class(self) -> None:
        if not is_native():
            self.skipTest("type parameters are only supported in the native parser")
        m, scopes = get_scope_metadata_provider(
            """
                class Outer:
                    class Nested:
                        pass
                    
                    type Alias = Nested

                    def meth1[T: Nested](self): pass
                    def meth2[T](self, arg: Nested): pass
            """
        )
        outer = ensure_type(m.body[0], cst.ClassDef)
        nested = ensure_type(outer.body.body[0], cst.ClassDef)
        alias = ensure_type(
            ensure_type(outer.body.body[1], cst.SimpleStatementLine).body[0],
            cst.TypeAlias,
        )
        self.assertIsInstance(scopes[alias.value], AnnotationScope)
        nested_refs_within_alias = list(scopes[alias.value].accesses["Nested"])
        self.assertEqual(len(nested_refs_within_alias), 1)
        self.assertEqual(
            {
                ensure_type(ref, Assignment).node
                for ref in nested_refs_within_alias[0].referents
            },
            {nested},
        )

        meth1 = ensure_type(outer.body.body[2], cst.FunctionDef)
        self.assertIsInstance(scopes[meth1], ClassScope)
        assert meth1.type_parameters
        meth1_typevar = ensure_type(meth1.type_parameters.params[0].param, cst.TypeVar)
        meth1_typevar_scope = scopes[meth1_typevar]
        self.assertIsInstance(meth1_typevar_scope, AnnotationScope)
        nested_refs_within_meth1 = list(meth1_typevar_scope.accesses["Nested"])
        self.assertEqual(len(nested_refs_within_meth1), 1)
        self.assertEqual(
            {
                ensure_type(ref, Assignment).node
                for ref in nested_refs_within_meth1[0].referents
            },
            {nested},
        )

        meth2 = ensure_type(outer.body.body[3], cst.FunctionDef)
        meth2_annotation = meth2.params.params[1].annotation
        assert meth2_annotation
        nested_refs_within_meth2 = list(scopes[meth2_annotation].accesses["Nested"])
        self.assertEqual(len(nested_refs_within_meth2), 1)
        self.assertEqual(
            {
                ensure_type(ref, Assignment).node
                for ref in nested_refs_within_meth2[0].referents
            },
            {nested},
        )

    def test_body_isnt_subject_to_special_annotation_rule(self) -> None:
        if not is_native():
            self.skipTest("type parameters are only supported in the native parser")
        m, scopes = get_scope_metadata_provider(
            """
            class Outer:
                class Inner: pass
                def f[T: Inner](self): Inner
            """
        )
        outer = ensure_type(m.body[0], cst.ClassDef)
        # note: this is different from global scope
        outer_scope = scopes[outer.body.body[0]]
        inner_assignment = list(outer_scope["Inner"])[0]
        self.assertEqual(len(inner_assignment.references), 1)
        f = ensure_type(outer.body.body[1], cst.FunctionDef)
        assert f.type_parameters
        T = ensure_type(f.type_parameters.params[0].param, cst.TypeVar)
        self.assertIs(list(inner_assignment.references)[0].node, T.bound)

        inner_in_func_body = ensure_type(f.body.body[0], cst.Expr)
        f_scope = scopes[inner_in_func_body]
        self.assertIn(inner_in_func_body.value, f_scope.accesses)
        self.assertEqual(list(f_scope.accesses)[0].referents, set())



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import Any, Callable

import libcst as cst
from libcst import parse_expression, parse_statement, PartialParserConfig
from libcst._nodes.tests.base import CSTNodeTest
from libcst.metadata import CodeRange
from libcst.testing.utils import data_provider


class SimpleCompTest(CSTNodeTest):
    @data_provider(
        [
            # simple GeneratorExp
            {
                "node": cst.GeneratorExp(
                    cst.Name("a"), cst.CompFor(target=cst.Name("b"), iter=cst.Name("c"))
                ),
                "code": "(a for b in c)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 13)),
            },
            # simple ListComp
            {
                "node": cst.ListComp(
                    cst.Name("a"), cst.CompFor(target=cst.Name("b"), iter=cst.Name("c"))
                ),
                "code": "[a for b in c]",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 0), (1, 14)),
            },
            # simple SetComp
            {
                "node": cst.SetComp(
                    cst.Name("a"), cst.CompFor(target=cst.Name("b"), iter=cst.Name("c"))
                ),
                "code": "{a for b in c}",
                "parser": parse_expression,
            },
            # non-trivial elt in GeneratorExp
            {
                "node": cst.GeneratorExp(
                    cst.BinaryOperation(cst.Name("a1"), cst.Add(), cst.Name("a2")),
                    cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                ),
                "code": "(a1 + a2 for b in c)",
                "parser": parse_expression,
            },
            # non-trivial elt in ListComp
            {
                "node": cst.ListComp(
                    cst.BinaryOperation(cst.Name("a1"), cst.Add(), cst.Name("a2")),
                    cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                ),
                "code": "[a1 + a2 for b in c]",
                "parser": parse_expression,
            },
            # non-trivial elt in SetComp
            {
                "node": cst.SetComp(
                    cst.BinaryOperation(cst.Name("a1"), cst.Add(), cst.Name("a2")),
                    cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                ),
                "code": "{a1 + a2 for b in c}",
                "parser": parse_expression,
            },
            # async GeneratorExp
            {
                "node": cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        asynchronous=cst.Asynchronous(),
                    ),
                ),
                "code": "(a async for b in c)",
                "parser": lambda code: parse_expression(
                    code, config=PartialParserConfig(python_version="3.7")
                ),
            },
            # Python 3.6 async GeneratorExp
            {
                "node": cst.FunctionDef(
                    cst.Name("foo"),
                    cst.Parameters(),
                    cst.IndentedBlock(
                        (
                            cst.SimpleStatementLine(
                                (
                                    cst.Expr(
                                        cst.GeneratorExp(
                                            cst.Name("a"),
                                            cst.CompFor(
                                                target=cst.Name("b"),
                                                iter=cst.Name("c"),
                                                asynchronous=cst.Asynchronous(),
                                            ),
                                        )
                                    ),
                                )
                            ),
                        )
                    ),
                    asynchronous=cst.Asynchronous(),
                ),
                "code": "async def foo():\n    (a async for b in c)\n",
                "parser": lambda code: parse_statement(
                    code, config=PartialParserConfig(python_version="3.6")
                ),
            },
            # a generator doesn't have to own it's own parenthesis
            {
                "node": cst.Call(
                    cst.Name("func"),
                    [
                        cst.Arg(
                            cst.GeneratorExp(
                                cst.Name("a"),
                                cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                                lpar=[],
                                rpar=[],
                            )
                        )
                    ],
                ),
                "code": "func(a for b in c)",
                "parser": parse_expression,
            },
            # add a few 'if' clauses
            {
                "node": cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        ifs=[
                            cst.CompIf(cst.Name("d")),
                            cst.CompIf(cst.Name("e")),
                            cst.CompIf(cst.Name("f")),
                        ],
                    ),
                ),
                "code": "(a for b in c if d if e if f)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 28)),
            },
            # nested/inner for-in clause
            {
                "node": cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        inner_for_in=cst.CompFor(
                            target=cst.Name("d"), iter=cst.Name("e")
                        ),
                    ),
                ),
                "code": "(a for b in c for d in e)",
                "parser": parse_expression,
            },
            # nested/inner for-in clause with an 'if' clause
            {
                "node": cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        ifs=[cst.CompIf(cst.Name("d"))],
                        inner_for_in=cst.CompFor(
                            target=cst.Name("e"), iter=cst.Name("f")
                        ),
                    ),
                ),
                "code": "(a for b in c if d for e in f)",
                "parser": parse_expression,
            },
            # custom whitespace
            {
                "node": cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        ifs=[
                            cst.CompIf(
                                cst.Name("d"),
                                whitespace_before=cst.SimpleWhitespace("\t"),
                                whitespace_before_test=cst.SimpleWhitespace("\t\t"),
                            )
                        ],
                        whitespace_before=cst.SimpleWhitespace("  "),
                        whitespace_after_for=cst.SimpleWhitespace("   "),
                        whitespace_before_in=cst.SimpleWhitespace("    "),
                        whitespace_after_in=cst.SimpleWhitespace("     "),
                    ),
                    lpar=[cst.LeftParen(whitespace_after=cst.SimpleWhitespace("\f"))],
                    rpar=[
                        cst.RightParen(whitespace_before=cst.SimpleWhitespace("\f\f"))
                    ],
                ),
                "code": "(\fa  for   b    in     c\tif\t\td\f\f)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 2), (1, 30)),
            },
            # custom whitespace around ListComp's brackets
            {
                "node": cst.ListComp(
                    cst.Name("a"),
                    cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                    lbracket=cst.LeftSquareBracket(
                        whitespace_after=cst.SimpleWhitespace("\t")
                    ),
                    rbracket=cst.RightSquareBracket(
                        whitespace_before=cst.SimpleWhitespace("\t\t")
                    ),
                    lpar=[cst.LeftParen(whitespace_after=cst.SimpleWhitespace("\f"))],
                    rpar=[
                        cst.RightParen(whitespace_before=cst.SimpleWhitespace("\f\f"))
                    ],
                ),
                "code": "(\f[\ta for b in c\t\t]\f\f)",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 2), (1, 19)),
            },
            # custom whitespace around SetComp's braces
            {
                "node": cst.SetComp(
                    cst.Name("a"),
                    cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                    lbrace=cst.LeftCurlyBrace(
                        whitespace_after=cst.SimpleWhitespace("\t")
                    ),
                    rbrace=cst.RightCurlyBrace(
                        whitespace_before=cst.SimpleWhitespace("\t\t")
                    ),
                    lpar=[cst.LeftParen(whitespace_after=cst.SimpleWhitespace("\f"))],
                    rpar=[
                        cst.RightParen(whitespace_before=cst.SimpleWhitespace("\f\f"))
                    ],
                ),
                "code": "(\f{\ta for b in c\t\t}\f\f)",
                "parser": parse_expression,
            },
            # no whitespace between elements
            {
                "node": cst.GeneratorExp(
                    cst.Name("a", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]),
                    cst.CompFor(
                        target=cst.Name(
                            "b", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
                        ),
                        iter=cst.Name(
                            "c", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
                        ),
                        ifs=[
                            cst.CompIf(
                                cst.Name(
                                    "d", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
                                ),
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_before_test=cst.SimpleWhitespace(""),
                            )
                        ],
                        inner_for_in=cst.CompFor(
                            target=cst.Name(
                                "e", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
                            ),
                            iter=cst.Name(
                                "f", lpar=[cst.LeftParen()], rpar=[cst.RightParen()]
                            ),
                            whitespace_before=cst.SimpleWhitespace(""),
                            whitespace_after_for=cst.SimpleWhitespace(""),
                            whitespace_before_in=cst.SimpleWhitespace(""),
                            whitespace_after_in=cst.SimpleWhitespace(""),
                        ),
                        whitespace_before=cst.SimpleWhitespace(""),
                        whitespace_after_for=cst.SimpleWhitespace(""),
                        whitespace_before_in=cst.SimpleWhitespace(""),
                        whitespace_after_in=cst.SimpleWhitespace(""),
                    ),
                    lpar=[cst.LeftParen()],
                    rpar=[cst.RightParen()],
                ),
                "code": "((a)for(b)in(c)if(d)for(e)in(f))",
                "parser": parse_expression,
                "expected_position": CodeRange((1, 1), (1, 31)),
            },
            # no whitespace before/after GeneratorExp is valid
            {
                "node": cst.Comparison(
                    cst.GeneratorExp(
                        cst.Name("a"),
                        cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                    ),
                    [
                        cst.ComparisonTarget(
                            cst.Is(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            cst.GeneratorExp(
                                cst.Name("d"),
                                cst.CompFor(target=cst.Name("e"), iter=cst.Name("f")),
                            ),
                        )
                    ],
                ),
                "code": "(a for b in c)is(d for e in f)",
                "parser": parse_expression,
            },
            # no whitespace before/after ListComp is valid
            {
                "node": cst.Comparison(
                    cst.ListComp(
                        cst.Name("a"),
                        cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                    ),
                    [
                        cst.ComparisonTarget(
                            cst.Is(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            cst.ListComp(
                                cst.Name("d"),
                                cst.CompFor(target=cst.Name("e"), iter=cst.Name("f")),
                            ),
                        )
                    ],
                ),
                "code": "[a for b in c]is[d for e in f]",
                "parser": parse_expression,
            },
            # no whitespace before/after SetComp is valid
            {
                "node": cst.Comparison(
                    cst.SetComp(
                        cst.Name("a"),
                        cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                    ),
                    [
                        cst.ComparisonTarget(
                            cst.Is(
                                whitespace_before=cst.SimpleWhitespace(""),
                                whitespace_after=cst.SimpleWhitespace(""),
                            ),
                            cst.SetComp(
                                cst.Name("d"),
                                cst.CompFor(target=cst.Name("e"), iter=cst.Name("f")),
                            ),
                        )
                    ],
                ),
                "code": "{a for b in c}is{d for e in f}",
                "parser": parse_expression,
            },
        ]
    )
    def test_valid(self, **kwargs: Any) -> None:
        self.validate_node(**kwargs)

    @data_provider(
        (
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                    lpar=[cst.LeftParen(), cst.LeftParen()],
                    rpar=[cst.RightParen()],
                ),
                "unbalanced parens",
            ),
            (
                lambda: cst.ListComp(
                    cst.Name("a"),
                    cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                    lpar=[cst.LeftParen(), cst.LeftParen()],
                    rpar=[cst.RightParen()],
                ),
                "unbalanced parens",
            ),
            (
                lambda: cst.SetComp(
                    cst.Name("a"),
                    cst.CompFor(target=cst.Name("b"), iter=cst.Name("c")),
                    lpar=[cst.LeftParen(), cst.LeftParen()],
                    rpar=[cst.RightParen()],
                ),
                "unbalanced parens",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        whitespace_before=cst.SimpleWhitespace(""),
                    ),
                ),
                "Must have at least one space before 'for' keyword.",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        asynchronous=cst.Asynchronous(),
                        whitespace_before=cst.SimpleWhitespace(""),
                    ),
                ),
                "Must have at least one space before 'async' keyword.",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        whitespace_after_for=cst.SimpleWhitespace(""),
                    ),
                ),
                "Must have at least one space after 'for' keyword.",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        whitespace_before_in=cst.SimpleWhitespace(""),
                    ),
                ),
                "Must have at least one space before 'in' keyword.",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        whitespace_after_in=cst.SimpleWhitespace(""),
                    ),
                ),
                "Must have at least one space after 'in' keyword.",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        ifs=[
                            cst.CompIf(
                                cst.Name("d"),
                                whitespace_before=cst.SimpleWhitespace(""),
                            )
                        ],
                    ),
                ),
                "Must have at least one space before 'if' keyword.",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        ifs=[
                            cst.CompIf(
                                cst.Name("d"),
                                whitespace_before_test=cst.SimpleWhitespace(""),
                            )
                        ],
                    ),
                ),
                "Must have at least one space after 'if' keyword.",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        inner_for_in=cst.CompFor(
                            target=cst.Name("d"),
                            iter=cst.Name("e"),
                            whitespace_before=cst.SimpleWhitespace(""),
                        ),
                    ),
                ),
                "Must have at least one space before 'for' keyword.",
            ),
            (
                lambda: cst.GeneratorExp(
                    cst.Name("a"),
                    cst.CompFor(
                        target=cst.Name("b"),
                        iter=cst.Name("c"),
                        inner_for_in=cst.CompFor(
                            target=cst.Name("d"),
                            iter=cst.Name("e"),
                            asynchronous=cst.Asynchronous(),
                            whitespace_before=cst.SimpleWhitespace(""),
                        ),
                    ),
                ),
                "Must have at least one space before 'async' keyword.",
            ),
        )
    )
    def test_invalid(
        self, get_node: Callable[[], cst.CSTNode], expected_re: str
    ) -> None:
        self.assert_invalid(get_node, expected_re)



# Copyright 2004-2005 Elemental Security, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.
#
# Modifications:
# Copyright David Halter and Contributors
# Modifications are dual-licensed: MIT and PSF.
# 99% of the code is different from pgen2, now.
#
# A fork of Parso's tokenize test
# https://github.com/davidhalter/parso/blob/master/test/test_tokenize.py
#
# The following changes were made:
# - Convert base test to Unittet
# - Remove grammar-specific tests
# pyre-unsafe
from libcst._parser.parso.python.tokenize import tokenize
from libcst._parser.parso.utils import parse_version_string
from libcst.testing.utils import data_provider, UnitTest


class ParsoTokenizeTest(UnitTest):
    @data_provider(
        (
            # 2 times 2, 5 because python expr and endmarker.
            ('f"}{"', [(1, 0), (1, 2), (1, 3), (1, 4), (1, 5)]),
            (
                'f" :{ 1 : } "',
                [
                    (1, 0),
                    (1, 2),
                    (1, 4),
                    (1, 6),
                    (1, 8),
                    (1, 9),
                    (1, 10),
                    (1, 11),
                    (1, 12),
                    (1, 13),
                ],
            ),
            (
                'f"""\n {\nfoo\n }"""',
                [(1, 0), (1, 4), (2, 1), (3, 0), (4, 1), (4, 2), (4, 5)],
            ),
        )
    )
    def test_tokenize_start_pos(self, code, positions):
        tokens = list(tokenize(code, version_info=parse_version_string("3.6")))
        assert positions == [p.start_pos for p in tokens]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Regenerate test fixtures, eg. after upgrading Pyre
"""

import json
import os
from pathlib import Path
from subprocess import run

from libcst.metadata import TypeInferenceProvider


def main() -> None:
    CWD = Path.cwd()
    repo_root = Path(__file__).parent.parent
    test_root = repo_root / "libcst" / "tests" / "pyre"

    try:
        os.chdir(test_root)
        run(["pyre", "-n", "start", "--no-watchman"], check=True)

        for file_path in test_root.glob("*.py"):
            json_path = file_path.with_suffix(".json")
            print(f"generating {file_path} -> {json_path}")

            path_str = file_path.as_posix()
            cache = TypeInferenceProvider.gen_cache(test_root, [path_str], timeout=None)
            result = cache[path_str]
            json_path.write_text(json.dumps(result, sort_keys=True, indent=2))

    finally:
        run(["pyre", "-n", "stop"], check=True)
        os.chdir(CWD)


if __name__ == "__main__":
    main()
